#!/usr/bin/env python

#
# This file is part of the `neko` Python module
#
# Copyright 2023
# Insitut Curie
#
# File author(s): Marco Ruscone (marco.ruscone@curie.fr)
#
# Distributed under the GPLv3 license
# See the file `LICENSE` or read a copy at
# https://www.gnu.org/licenses/gpl-3.0.txt
#

"""
Brief description of the project/module.
"""

__all__ = [
    '__version__',
    '__author__',
]

from ._metadata import __author__, __version__



import pandas as pd

def compare_networks(network1, network2):

    """
    This function compares two networks and identifies common, unique, and conflicting interactions and nodes.
    Args:
        network1:
        network2:

    Returns:

    """

    df1 = network1.convert_edgelist_into_genesymbol()
    df2 = network2.convert_edgelist_into_genesymbol()
    # Merge dataframes on 'source' and 'target' to identify common and conflicting interactions
    merged = pd.merge(df1, df2, on=['source', 'target'], suffixes=('_1', '_2'), how='outer', indicator=True)

    # Determine the comparison type for interactions
    merged['comparison'] = merged['_merge'].apply(
        lambda x: 'Unique to Network 1' if x == 'left_only' else ('Unique to Network 2' if x == 'right_only' else 'Common')
    )

    # Convert the comparison column to categorical and add the new category
    merged['comparison'] = pd.Categorical(merged['comparison'], categories=['Unique to Network 1', 'Unique to Network 2', 'Common', 'Conflicting'])

    # Update the comparison type for conflicting interactions
    conflicting_mask = (merged['_merge'] == 'both') & (merged['Effect_1'] != merged['Effect_2'])
    merged.loc[conflicting_mask, 'comparison'] = 'Conflicting'

    # Create the interaction comparison DataFrame
    interaction_comparison = merged[['source', 'target', 'comparison']]

    # Determine unique and common nodes
    nodes_1 = set(df1['source']).union(set(df1['target']))
    nodes_2 = set(df2['source']).union(set(df2['target']))

    unique_nodes_network_1 = nodes_1 - nodes_2
    unique_nodes_network_2 = nodes_2 - nodes_1
    common_nodes = nodes_1 & nodes_2

    # Create a list of node comparison data
    node_comparison_data = []
    for node in unique_nodes_network_1:
        node_comparison_data.append([node, 'Unique to Network 1'])
    for node in unique_nodes_network_2:
        node_comparison_data.append([node, 'Unique to Network 2'])
    for node in common_nodes:
        node_comparison_data.append([node, 'Common'])

    # Create the node comparison DataFrame
    node_comparison = pd.DataFrame(node_comparison_data, columns=['node', 'comparison'])

    return interaction_comparison, node_comparison



from __future__ import annotations

from typing import Any, Callable, Iterable, Literal
import pickle
import logging
import pandas as pd
import os
from pypath_common import misc as _common

from ._db.omnipath import omnipath_universe
from ._db import psp as _psp
from ._db import _misc
from ._db import huri as _huri
from ._db import signor as _signor

"""
Access to generic networks from databases, files and standard formats.
"""

_METHODS = {
    'omnipath': omnipath_universe,
}
_REQUIRED_COLS = {
    'source',
    'target',
}

MANDATORY_BOOL_COLS = [
    'is_directed',
    'is_stimulation',
    'is_inhibition',
    'form_complex',
]

MANDATORY_COLUMNS = [
    'source',
    'target',
]
MANDATORY_COLUMNS.extend(MANDATORY_BOOL_COLS)


def network_universe(
        resource: Literal['omnipath'] | pd.DataFrame = 'omnipath',
        **kwargs
    ) -> Universe:
    """
    Generic networks from databases, files and standard formats.

    Args:
        resource:
            Name of the resource or a ready data frame to bypass the built-in
            loading method.
        kwargs:
            Passed to the source specific method. See the specific methods in
            this module for details.

    Note: currently OmniPath PPI is the single available option and serves
    as a placeholder. Later we will dispatch all inputs through this API.
    """

    return Universe(resource, **kwargs)


def omnipath(**kwargs) -> Universe:

    return network_universe('omnipath', **kwargs)


def signor(path: str | None = None, **kwargs) -> Universe:

    if path and os.path.exists(path):

        return Universe(_signor.signor(path))

    else:

        raise NotImplementedError(
            'Direct retrieval of SIGNOR is not implemented. '
            'Please download manually the TSV file and provide the path.'
        )


def phosphosite(
        organism: Literal["human", "mouse", "rat"] = 'human',
        kinase_substrate: str | None = None,
        regulatory_sites: str | None = None,
        **kwargs
    ) -> Universe:

    df = _psp.psp(organism, kinase_substrate, regulatory_sites)

    return Universe(df, name = 'phosphosite')


def huri(dataset: str = 'HI-union') -> Universe:

    df = _huri.huri(dataset)

    return Universe(df, directed = False, name = 'huri')


class Universe:


    def __init__(
            self,
            resources: Literal['omnipath'] | pd.DataFrame = None,
            **param
        ):
        """
        Load and preprocess a generic network from databases or files.
        """

        self._resources = {}
        self._directed = {}
        self.interactions = None
        self.add_resources(resources, **param)
        self.build()


    def add_resources(self, resources, directed: bool = True, **param) -> None:

        if isinstance(resources, str):

            if resources.endswith('.pickle'):

                with open(resources, 'rb') as fin:

                    resources = pickle.load(fin)

            elif resources.endswith('.tsv'):

                resources = pd.read_csv(resources, sep='\t')

        name = param.get('name', '_default')
        columns = param.get('columns', {})

        if isinstance(resources, str) and resources in _METHODS:

            name = name or resources
            resources = _METHODS[resources](**param)

        if isinstance(resources, pd.DataFrame):

            self._resources[name] = self._check_columns(
                resources,
                columns,
                directed = directed,
            )

        elif isinstance(resources, dict):

            resources = {
                k: self._check_columns(v, columns)
                for k, v in resources.items()
            }

            if isinstance(directed, bool):

                directed = {k: directed for k in resources.keys()}

            for name, resource in resources.items():

                self.add_resources(
                    resource,
                    directed = directed[name],
                    name = name,
                    **param,
                )

        elif isinstance(resources, Iterable):

            names = [f'{name}_{i}' for i in range(len(resources))]

            if isinstance(directed, bool):

                directed = [directed] * len(resources)

            for r, d, n in zip(resources, directed, names):

                self.add_resources(r, directed = d, name = n, **param)


    @staticmethod
    def _check_columns(
            df: pd.DataFrame,
            columns: dict,
            directed: bool = True,
        ) -> pd.DataFrame:

        # If columns is provided, rename the columns of the incoming df
        if columns:
            df = df.rename(columns=columns)

        if 'effect' in df.columns:

            df = _misc.split_effect(df)

        for col in MANDATORY_BOOL_COLS:

            df = _misc.bool_col(df, col)

        # Check if the df contains the required columns
        missing_columns = set(MANDATORY_COLUMNS) - set(df.columns)

        if missing_columns:

            logging.warning("The incoming df is missing some required columns: %s", missing_columns)
            logging.warning("This might lead to issues in running the package.")

        if not directed:

            df = _misc.undirected_to_mutual(df)

        return df


    @staticmethod
    def merge(df1: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
        """
        This function concatenates the provided df with the existing one in the resources object,
        aligning columns and filling in missing data with NaN.

        Parameters:
            df1 (pd.DataFrame): The DataFrame to be added.
            df2 (pd.DataFrame): The DataFrame to be added.

        Raises:
            ValueError: If the 'df' parameter is not a pandas DataFrame.

        Returns:
            None
        """

        # Align columns of both dataframes, filling missing columns with NaN
        all_columns = set(df1.columns).union(set(df2.columns))
        df1 = df1.reindex(columns=all_columns, fill_value=None)
        df2 = df2.reindex(columns=all_columns, fill_value=None)
        df1 = pd.concat([df1, df2])

        return df1.copy()


    def build(self, resources: Iterable[str] | None = None) -> None:

        resources = _common.to_list(resources) or self._resources.keys()

        self.interactions = None

        if not resources:

            return

        for res in resources:
            df = self._resources[res]
            self.interactions = (
                df.copy()
                if self.interactions is None else
                self.merge(self.interactions, df)
            )

        self.interactions.reset_index(drop=True, inplace=True)


    @property
    def resource(self):

        return self._resource if isinstance(self._resource, str) else 'user'


    @property
    def network(self) -> pd.DataFrame:
        """
        The network as it's been read from the original source.
        """

        if not hasattr(self, '_network'):

            self.load()

        return self._network


    @network.setter
    def network(self, value: Any):

       raise AttributeError('The attribute `Universe.network` is read-only.')


    def load(self) -> None:
        """
        Acquire the input data according to parameters.
        """

        self._network = self.method(**self.param)


    @property
    def method(self) -> Callable:
        """
        The method that loads the data.

        ``param`` are to be passed to this method.
        """

        return (
            lambda **kwargs: self._resource
                if isinstance(self._resource, pd.DataFrame) else
            _METHODS[self.resource]
        )


    def __repr__(self) -> str:

        return f'Universe; resources: {", ".join(self._resources.keys())}; size: {len(self)}'


    def __len__(self) -> int:

        return 0 if self.interactions is None else len(self.interactions)


    def check(self) -> bool:
        """
        The network is loaded and contains the mandatory variables.
        """

        return (
            hasattr(self, '_network') and
            not _REQUIRED_COLS - set(self._network.columns)
        )

    @property
    def nodes(self) -> set[str]:

        return (
            set(self.interactions['source']) |
            set(self.interactions['target'])
        )


    def __contains__(self, other) -> bool:

        return other in self.nodes


    def __and__(self, other: set) -> set:

        return self.nodes & other



from __future__ import annotations
from typing import List, Optional
from pypath.utils import mapping
from itertools import combinations
import pandas as pd
from ..inputs import _universe
from .._methods.enrichment_methods import Connections
from typing_extensions import Literal
import copy
from .._annotations.gene_ontology import Ontology

import networkx as nx


def is_connected(network) -> bool:
    """
    This function checks if a network is connected. It takes a Network object as input and returns True if the network
    is connected, otherwise it returns False.

    Args:
        - network: A Network object representing the network to be checked.

    Returns:
        - bool
    """
    # Create a graph from the edges
    g = nx.from_pandas_edgelist(network.edges, 'source', 'target')
    # Add isolated nodes to the graph
    all_nodes = set(network.nodes['Uniprot'])
    g.add_nodes_from(all_nodes)
    # Check if the graph is connected
    return nx.is_connected(g)


def check_sign(interaction: pd.DataFrame, consensus: bool = False) -> str:
    """
    This function checks the sign of an interaction in the Omnipath format (Pandas DataFrame or Series).
    The attribute "consensus" checks for the consistency of the sign of the interaction among the references.

    Args:
        - interaction: A pandas DataFrame or Series representing the interaction.
        - consensus: A boolean indicating whether to check for consensus among references.

    Returns:
        - A string indicating the sign of the interaction: "stimulation", "inhibition", "form complex", or "undefined".
    """
    # Handle both DataFrame and Series input
    if isinstance(interaction, pd.DataFrame):
        interaction = interaction.iloc[0]

    if consensus:
        if interaction.get("consensus_inhibition") and interaction.get("consensus_stimulation"):
            return "bimodal"
        if interaction.get("consensus_stimulation"):
            return "stimulation"
        elif interaction.get("consensus_inhibition"):
            return "inhibition"
        else:
            return "undefined"
    else:
        # Check if it is both stimulation and inhibition
        if interaction.get("is_stimulation", True) and interaction.get("is_inhibition", True):
            return "bimodal"
        if interaction.get("is_stimulation", False):
            return "stimulation"
        elif interaction.get("is_inhibition", False):
            return "inhibition"
        # Check for "form_complex" column existence
        elif interaction.get("form_complex", False):
            return "form complex"
        else:
            return "undefined"


def check_gene_list_format(gene_list: list[str]) -> bool:
    """
    This function checks the format of the gene list and returns True if the gene list is in Uniprot format,
    False if the gene list is in genesymbol format.

    Args:
        - gene_list: A list of gene identifiers. The gene identifiers can be either Uniprot identifiers or genesymbols.

    Returns:
        - A boolean indicating whether the gene list is in Uniprot format (True) or genesymbol format (False).
    """
    # Check if the gene list contains Uniprot identifiers
    if all(mapping.id_from_label0(gene) for gene in gene_list):
        return True
    # Check if the gene list contains genesymbols
    elif all(mapping.label(gene) for gene in gene_list):
        return False


def mapping_node_identifier(node: str) -> list[str]:
    """
    This function takes a node identifier and returns a list containing the possible identifiers for the node.
    The identifiers include a complex string, a genesymbol, and a uniprot identifier. The function uses the
    mapping.id_from_label0 and mapping.label functions from the pypath.utils.mapping module to translate the node
    identifier into these different formats.

    Args:
        - node: A string representing the node identifier. The node identifier can be a genesymbol, a uniprot identifier,
                or a complex string.

    Returns:
        - A list containing the complex string, genesymbol, and uniprot identifier for the node. If the node identifier
          cannot be translated into one of these formats, the corresponding value in the list is None.
    """
    complex_string = None
    genesymbol = None
    uniprot = None

    if mapping.id_from_label0(node):
        # Convert UniProt ID to gene symbol
        uniprot = mapping.id_from_label0(node)
        if uniprot.startswith("MI"):
            genesymbol = uniprot
        else:
            # Set the UniProt ID as the 'Uniprot' value in the new entry
            genesymbol = mapping.label(uniprot)
    elif mapping.id_from_label0(node).startswith("COMPLEX"):
        node = node[8:]
        node_list = node.split("_")

        # Translate each element in node_list using mapping.label
        translated_node_list = [mapping.label(mapping.id_from_label0(item)) for item in node_list]

        # Join the elements in node_list with "_"
        joined_node_string = "_".join(translated_node_list)

        # Add back the "COMPLEX:" prefix to the string
        complex_string = "COMPLEX:" + joined_node_string
    elif mapping.label(node):
        genesymbol = mapping.label(node)
        uniprot = mapping.id_from_label0(genesymbol)
    else:
        print("Error during translation, check syntax for ", node)

    return [complex_string, genesymbol, uniprot]


def translate_paths(paths) -> list[list[str]]:
    """
    This function translates a list of paths, where each path is a sequence of node identifiers.
    It uses the helper function `handle_complex_identifier` to translate each node identifier in the paths.

    Args:
        - paths: A list of paths, where each path is a sequence of node identifiers.
                 A node identifier can be a string or a list of strings.

    Returns:
        - A list of translated paths, where each path is a sequence of translated node identifiers.
    """
    translated_list = []

    def handle_complex_identifier(item):
        """
        This helper function translates a node identifier using the `mapping_node_identifier` function.
        It checks all possible identifiers (complex, genesymbol, uniprot) and returns the first non-None value.

        Args:
        - item: A node identifier.

        Returns:
        - The translated node identifier.
        """
        identifiers = mapping_node_identifier(item)
        return identifiers[0] or identifiers[1] or identifiers[2]

    # If input_list is a list of strings
    if isinstance(paths[0], str):
        translated_list = [handle_complex_identifier(item) for item in paths]
    # If input_list is a list of lists of strings
    elif isinstance(paths[0], list):
        for sublist in paths:
            translated_sublist = [handle_complex_identifier(item) for item in sublist]
            translated_list.append(translated_sublist)

    return translated_list


def join_unique(series) -> str:
    """
    This function takes a pandas Series, filters out None values, and returns a string of unique values joined by a comma.

    Args:
        - series: A pandas Series object.

    Returns: - A string of unique values in the series, joined by a comma. If a value in the series is None,
                it is not included in the output string.
    """
    # Filter out None values before converting to set and joining
    filtered_series = [str(item) for item in series if item is not None]
    unique_items = set(filtered_series)
    return ', '.join(unique_items)


def determine_most_frequent_effect(effects):
    effect_counts = effects.value_counts()

    # If there's only one type of effect, return it
    if len(effect_counts) == 1:
        return effect_counts.index[0]

    # Count the occurrences of each effect type
    stimulation_count = effect_counts.get('stimulation', 0)
    inhibition_count = effect_counts.get('inhibition', 0)
    form_complex_count = effect_counts.get('form_complex', 0)

    # Calculate the total of stimulation, inhibition, and form_complex
    total_known_effects = stimulation_count + inhibition_count + form_complex_count

    # If there are more unknown effects than known ones, return 'undefined'
    if len(effects) - total_known_effects > total_known_effects:
        return 'undefined'

    # If form_complex is the most frequent, return it
    if form_complex_count > stimulation_count and form_complex_count > inhibition_count:
        return 'form_complex'

    # Handle stimulation and inhibition
    if stimulation_count > inhibition_count:
        return 'stimulation'
    elif inhibition_count > stimulation_count:
        return 'inhibition'
    elif stimulation_count == inhibition_count and stimulation_count > 0:
        return 'bimodal'

    # If we've reached this point, it means there's no clear majority
    return 'undefined'


class Network:
    """
    A molecular interaction network.

        The `Network` object is the central organizing component of the `neko`
        module. It is the subject of all operations implemented here, including
        topological algorithms, graph analysis, network visualization and
        integration of database knowledge.

    Args:
        initial_nodes: A list of initial nodes to be added to the network.
        sif_file: A SIF (Simple Interaction Format) file to load the network from.
        resources: A pandas DataFrame containing the resources database.

    Methods:
    """

    def __init__(
            self,
            initial_nodes: list[str] = None,
            sif_file=None,
            resources=None,
        ):

        self._init_args = locals()
        del self._init_args['self']
        self.nodes = pd.DataFrame(columns=["Genesymbol", "Uniprot", "Type"])
        self.edges = pd.DataFrame(columns=["source", "target", "Type", "Effect", "References"])
        self.initial_nodes = initial_nodes
        self.__ontology = Ontology()
        self._populate()


    def _populate(self):

        self.resources = (
            _universe.
            network_universe(self._init_args['resources']).
            interactions
        )

        if self.initial_nodes:
            for node in self.initial_nodes:
                self.add_node(node)
            self.__drop_missing_nodes()
            self.nodes.reset_index(inplace=True, drop=True)

        elif sif_files := self._init_args['sif_file']:
            self.initial_nodes = []
            self.__load_network_from_sif(sif_file)

        self.__connect = Connections(self.resources)
        self.__algorithms = {
            'dfs': self.dfs_algorithm,
            'bfs': self.bfs_algorithm
        }

    def copy(self):
        new_instance = copy.deepcopy(self)
        return new_instance

    def check_nodes(self, nodes: list[str]) -> list[str]:
        """
        This function checks if the nodes exist in the resources database and returns the nodes that are present.

        Args:
            - nodes: A list of node identifiers (strings). These are the nodes to be checked.

        Returns:
            - A list[str] of node identifiers that are present in the resources database.
        """
        return [node for node in nodes if
                node in self.resources["source"].unique() or node in self.resources["target"].unique()]

    def check_node(self, node: str) -> bool:
        """
        This function checks if a node exists in the resources' database.

        Args:
            - node: A string representing the node to be checked.

        Returns:
            - A boolean indicating whether the node exists in the resources' database.
        """
        return True if self.check_nodes([node]) else False

    def __drop_missing_nodes(self) -> None:
        """
        This function drops the nodes that are not present in the resources database and print a warning with the
        name of the missing nodes.

        The function works as follows:
        1. It first calls the `check_nodes` function to get a list of nodes that exist in the resources' database.
        2. It then finds the nodes in the network that are not in this list, and removes them from the network.
        3. If there are any missing nodes, it prints a warning with their names.

        This function does not return anything. It modifies the `nodes` attribute of the `Network` object in-place.
        """
        # Get the list of nodes that exist in the resources database
        existing_nodes = list(set(self.check_nodes(self.nodes["Uniprot"].tolist()))
                              |
                              set(self.check_nodes(self.nodes["Genesymbol"].tolist())))

        # Create a mask for nodes that exist in either Uniprot or Genesymbol lists
        existing_mask = self.nodes["Uniprot"].isin(existing_nodes) | self.nodes["Genesymbol"].isin(existing_nodes)
        # Identify removed nodes
        removed_nodes = self.nodes[~existing_mask]

        # Print a warning with the name of the missing nodes
        if not removed_nodes.empty:
            missing_uniprot = removed_nodes["Uniprot"].tolist()
            missing_genesymbol = removed_nodes["Genesymbol"].tolist()
            missing_identifiers = list(set(missing_uniprot + missing_genesymbol))  # Remove duplicates
            print(
                "Warning: The following nodes were not found in the resources database and have been removed from the "
                "network:",
                ", ".join(str(node) for node in missing_identifiers if pd.notna(node))
            )
        # Keep only the nodes that exist in the resources database
        self.nodes = self.nodes[existing_mask]

        return

    def add_node(self, node: str, from_sif: bool = False) -> None:
        """
        Adds a node to the network. The node is added to the nodes DataFrame of the network. The function checks the
        syntax for the genesymbol to ensure it is correct. If the node is a complex, it is added with the
        'Genesymbol' as the complex string and 'Uniprot' as the node. Otherwise, it is added with the 'Genesymbol' as
        the genesymbol and 'Uniprot' as the uniprot. The 'Type' is set as 'NaN' for all new nodes.

        Args:
            - node: A string representing the node to be added. The node can be represented by either its
                    Genesymbol or Uniprot identifier.

        Returns:
            - None.
        """

        if from_sif:
            # check that the new entry node can be translated using the function mapping node identifier (all the
            # output of the function should be None) if it cannot be translated, print an error message but add the
            # node to the network anyway

            complex_string, genesymbol, uniprot = mapping_node_identifier(node)
            if not complex_string and not genesymbol and not uniprot:
                print("Error: node %s could not be automatically translated" % node)
                new_entry = {"Genesymbol": node, "Uniprot": node, "Type": "NaN"}
                self.nodes.loc[len(self.nodes)] = new_entry
                self.nodes = self.nodes.drop_duplicates().reset_index(drop=True)

            new_entry = {"Genesymbol": genesymbol, "Uniprot": uniprot, "Type": "NaN"}
            self.nodes.loc[len(self.nodes)] = new_entry
            self.nodes = self.nodes.drop_duplicates().reset_index(drop=True)
            self.initial_nodes.append(new_entry["Genesymbol"])
            self.initial_nodes = list(set(self.initial_nodes))
            return

        complex_string, genesymbol, uniprot = mapping_node_identifier(node)

        if complex_string:
            new_entry = {"Genesymbol": complex_string, "Uniprot": node, "Type": "NaN"}
        else:
            new_entry = {"Genesymbol": genesymbol, "Uniprot": uniprot, "Type": "NaN"}

        if not self.check_node(uniprot) and not self.check_node(genesymbol):
            print("Error: node %s is not present in the resources database" % node)
            return

        self.nodes.loc[len(self.nodes)] = new_entry

        self.nodes = self.nodes.drop_duplicates().reset_index(drop=True)
        return

    def remove_node(self, node: str) -> None:
        """
        Removes a node from the network. The node is removed from both the list of nodes and the list of edges.

        Args:
            - node: A string representing the node to be removed. The node can be represented by either its
                    Genesymbol or Uniprot identifier.

        Returns:
            - None
        """
        # Remove the node from the nodes DataFrame
        self.nodes = self.nodes[(self.nodes.Genesymbol != node) & (self.nodes.Uniprot != node)]

        # Translate the node identifier to Uniprot
        node = mapping_node_identifier(node)[2]

        # Remove any edges associated with the node from the edges DataFrame
        self.edges = self.edges[~self.edges[['source', 'target']].isin([node]).any(axis=1)]

        return

    def add_edge(self, edge: pd.DataFrame) -> None:
        """
        This method adds an interaction to the list of interactions while converting it to the NeKo-network format.
        It checks if the edge represents inhibition or stimulation and sets the effect accordingly. It also checks if the
        nodes involved in the interaction are already present in the network, if not, it adds them.

        Args:
            - edge: A pandas DataFrame representing the interaction. The DataFrame should contain columns for
            'source', 'target', 'type', and 'references'. The 'source' and 'target' columns represent the nodes involved
            in the interaction. The 'type' column represents the type of interaction. The 'references' column contains
            the references for the interaction.

        Returns:
            - None
        """

        # Check if the edge represents inhibition or stimulation and set the effect accordingly
        effect = check_sign(edge)
        # check if the column reference is present in the edge dataframe
        references = edge["references"].values[0] if "references" in edge.columns else None
        # Get the type value from the edge DataFrame or set it to None
        edge_type = edge["type"].values[0] if "type" in edge.columns else None

        # Create a new DataFrame with edge information, including handling None for type
        df_edge = pd.DataFrame({
            "source": edge["source"],
            "target": edge["target"],
            "Type": edge_type,
            "Effect": effect,
            "References": references
        })

        # Convert the "Uniprot" column to a set for efficient membership test
        uniprot_nodes = set(self.nodes["Uniprot"].unique())

        # add the new nodes to the nodes dataframe
        if edge["source"].values[0] not in uniprot_nodes:
            self.add_node(edge["source"].values[0])
        if edge["target"].values[0] not in uniprot_nodes:
            self.add_node(edge["target"].values[0])

        # if in the edge dataframe there is an edge with the same source, target and effect, merge the references
        existing_edge = self.edges[(self.edges["source"] == edge["source"].values[0]) &
                                   (self.edges["target"] == edge["target"].values[0]) &
                                   (self.edges["Effect"] == effect)]
        if not existing_edge.empty and references is not None:
            self.edges.loc[existing_edge.index, "References"] += "; " + str(references)
        else:
            # Concatenate the new edge DataFrame with the existing edges in the graph
            self.edges = pd.concat([self.edges, df_edge])

        self.edges = self.edges.drop_duplicates().reset_index(drop=True)
        return

    def remove_edge(self, node1: str, node2: str) -> None:
        """
        This function removes an edge from the network. It takes the source node and target node as input and removes
        the edge from the edges DataFrame.

        Args:
            - node1: A string representing the source node of the edge.
            - node2: A string representing the target node of the edge.

        Returns:
            - None
        """
        # check if node1 and node2 are in genesymbol format or uniprot format
        if check_gene_list_format([node1]):
            node1 = mapping_node_identifier(node1)[2]
        if check_gene_list_format([node2]):
            node2 = mapping_node_identifier(node2)[2]

        # Remove the edge from the edges DataFrame, if the effect or the nodes are not present, print a warning
        if not self.edges[(self.edges["source"] == node1) & (self.edges["target"] == node2)].empty:
            self.edges = self.edges[~((self.edges["source"] == node1) & (self.edges["target"] == node2))]
        else:
            print("Warning: The edge does not exist in the network, check syntax for ",
                  mapping_node_identifier(node1)[1], " and ", mapping_node_identifier(node2)[1])
        return

    def remove_disconnected_nodes(self) -> None:
        """
        This function removes nodes from the network that are not connected to any other nodes.

        Returns:
        None. The function modifies the network object in-place by removing the disconnected nodes from the nodes DataFrame.
        """
        # Get the list of nodes that are not connected to any other nodes
        disconnected_nodes = self.nodes[~self.nodes["Uniprot"].isin(self.edges["source"]) &
                                        ~self.nodes["Uniprot"].isin(self.edges["target"])]

        # Remove the disconnected nodes from the nodes DataFrame
        self.nodes = self.nodes[~self.nodes["Uniprot"].isin(disconnected_nodes["Uniprot"])]

        return

    def modify_node_name(self, old_name: str, new_name: str,
                         type: Literal['Genesymbol', 'Uniprot', 'both'] = 'Genesymbol'
                         ) -> None:
        """
        This function modifies the name of a node in the network. It takes the old name of the node and the new name
        as input and modifies the name of the node in the nodes and in the edges DataFrame. If type is set to
        'Genesymbol', it modifies the genesymbol name of the node in the nodes DataFrame. If type is set to
        'Uniprot', it modifies the uniprot name of the node in the edges DataFrame. If type is set to 'both',
        it modifies both the genesymbol and uniprot names of the node in the nodes and edges DataFrame.


        Args:
            - old_name: A string representing the old name of the node. - new_name: A string representing the new
            name of the node. - type: A string indicating the type of name to be modified. It can be 'Genesymbol',
            'Uniprot', or 'both'. Default is 'Genesymbol'.

        Returns:
            -None
        """

        if type == 'Genesymbol':
            self.nodes.loc[self.nodes["Genesymbol"] == old_name, "Genesymbol"] = new_name
        elif type == 'Uniprot':
            self.nodes.loc[self.nodes["Uniprot"] == old_name, "Uniprot"] = new_name
            # Update the source and target columns in the edges DataFrame
            self.edges.loc[self.edges["source"] == old_name, "source"] = new_name
            self.edges.loc[self.edges["target"] == old_name, "target"] = new_name
        elif type == 'both':
            self.nodes.loc[self.nodes["Genesymbol"] == old_name, "Genesymbol"] = new_name
            # check if it is possible to translate the genesymbol to uniprot
            try:
                new_name_uniprot = mapping_node_identifier(new_name)[2]
                old_name_uniprot = mapping_node_identifier(old_name)[2]
            except:
                new_name_uniprot = new_name
                old_name_uniprot = old_name
            self.nodes.loc[self.nodes["Uniprot"] == old_name_uniprot, "Uniprot"] = new_name_uniprot
            # Update the source and target columns in the edges DataFrame
            self.edges.loc[self.edges["source"] == old_name_uniprot, "source"] = new_name_uniprot
            self.edges.loc[self.edges["target"] == old_name_uniprot, "target"] = new_name_uniprot
        else:
            print("Error: Invalid type. Please choose 'Genesymbol', 'Uniprot', or 'both'.")

        return

    def print_my_paths(self,
                       node1: str,
                       node2: str,
                       maxlen: int = 2,
                       genesymbol: bool = True
                       ) -> None:
        """
        This function prints all the paths between two nodes in the network. It uses the `find_paths` method from the
        `Connections` class to find all the paths between the two nodes in the Network object. If no paths are found,
        it prints a warning message. If one of the selected nodes is not present in the network, it prints an error message.

        Args:
            - node1: A string representing the source node.
            - node2: A string representing the target node.
            - maxlen: An integer representing the maximum length of the paths to be searched for. Default is 2.
            - genesymbol: A boolean flag indicating whether to print the paths in genesymbol format. Default is True.

        Returns:
            - None
        """

        # check if node1 and node2 are in genesymbol format or uniprot format
        if check_gene_list_format([node1]):
            node1 = mapping_node_identifier(node1)[2]
        if check_gene_list_format([node2]):
            node2 = mapping_node_identifier(node2)[2]

        # Check if the nodes exist in the network
        if node1 not in self.nodes["Uniprot"].tolist() or node2 not in self.nodes["Uniprot"].tolist():
            print("Error: One or both of the selected nodes are not present in the network.")
            return
        connect = Connections(self.edges)
        paths = connect.find_paths(node1, node2, maxlen=maxlen)

        if not paths:
            print("Warning: No paths found between source: ", node1, " and target: ", node2)
            return

        if genesymbol:
            paths = translate_paths(paths)

        # Print all the paths
        for path in paths:
            print(path)

        return

    def remove_path(self, path: list[str]) -> None:
        """
        This function removes a path from the network. It takes a list of nodes representing the path and removes all
        the edges between the nodes in the path.

        Args:
            - path: A list of nodes representing the path to be removed. The nodes can be represented as strings or tuples.

        Returns:
            - None
        """
        # check if node1 and node2 are in genesymbol format or uniprot format
        if check_gene_list_format(path):
            path = [mapping_node_identifier(node)[2] for node in path]

        # Iterate through the nodes in the path and remove the edges between them
        for i in range(0, len(path)):
            if i == len(path) - 1:
                break
            self.remove_edge(path[i], path[i + 1])
        return

    def __load_network_from_sif(self, sif_file) -> None:
        """
        Load a network object from a SIF (Simple Interaction Format) file.

        Args:
            - sif_file: A string representing the path to the SIF file.

        Returns:
            None
        """
        interactions = []
        node_set = set()

        def determine_effect(interaction_type):
            """
            Determine the effect based on the interaction type.

            Args:
            - interaction_type: A string representing the type of interaction.

            Returns:
            - A string representing the effect of the interaction. If the interaction type is not recognized, it returns "undefined".
            """
            effect_types = {
                "1": "stimulation",
                "activate": "stimulation",
                "stimulate": "stimulation",
                "phosphorilate": "undefined",
                "stimulation": "stimulation",
                "->": "stimulation",
                "-|": "inhibition",
                "-1": "inhibition",
                "inhibit": "inhibition",
                "block": "inhibition",
                "inhibition": "inhibition",
                "form_complex": "form complex",
                "form-complex": "form complex",
                "complex_formation": "form complex",
                "bimodal": "bimodal",
                "both": "bimodal"
            }
            return effect_types.get(interaction_type, "undefined")

        with open(sif_file, "r") as f:
            for line in f:
                if line.startswith('#'):  # Skip comment lines
                    continue
                interaction = line.strip().split()
                if len(interaction) < 3:
                    continue  # Skip malformed lines

                effect = determine_effect(interaction[1])

                interactions.append({
                    "source": mapping_node_identifier(interaction[0])[2] if check_gene_list_format(
                        [interaction[0]]) else interaction[0],
                    "target": mapping_node_identifier(interaction[2])[2] if check_gene_list_format(
                        [interaction[2]]) else interaction[2],
                    "Type": interaction[3] if len(interaction) > 3 else None,
                    "Effect": effect,
                    "References": "SIF file"
                })
                node_set.update([interaction[0], interaction[2]])

        # Create or update the edges DataFrame
        df_edge = pd.DataFrame(interactions)
        self.edges = pd.concat([self.edges, df_edge], ignore_index=True)

        # Update the nodes list
        nodes = list(node_set)
        for node in nodes:
            self.add_node(node, from_sif=True)

        return

    def __add_paths_to_edge_list(self, paths) -> None:
        """
        This method adds paths to the edge list of the network. A path is a sequence of nodes where each node is
        connected to the next node in the sequence. The function checks if there is an interaction between each pair
        of nodes in the path in the resources' database. If an interaction exists, it is added to the edge list of
        the network.

        Args:
            - paths: A list of paths, where each path is a sequence of nodes. A node can be a string or a tuple.

        Returns:
            - None
        """
        # Access the resources database
        database = self.resources

        # Iterate through the list of paths
        for path in paths:
            # Handle single string or tuple
            if isinstance(path, (str)):
                path = [path]

            # Iterate through the nodes in the path
            for i in range(0, len(path)):
                # Break the loop if it's the last node in the path
                if i == len(path) - 1:
                    break

                # Check if there is an interaction between the current node and the next node in the resources database
                interaction = database.loc[(database["source"] == path[i]) &
                                           (database["target"] == path[i + 1])]

                # If an interaction exists, add it to the edge list of the network
                if not interaction.empty:
                    if not ((self.edges['source'] == interaction['source'].values[0]) &
                            (self.edges['target'] == interaction['target'].values[0])).any():
                        self.add_edge(interaction)

        # Remove duplicate edges from the edge list
        self.edges = self.edges.drop_duplicates().reset_index(drop=True)

        return

    def __add_cascade_to_edge_list(self, cascades) -> None:
        """
        This function adds cascades to the edge list of the network. A cascade is a sequence of nodes where each node is
        connected to the next node in the sequence. The function checks if there is an interaction between each pair of nodes
        in the cascade in the resources' database. If an interaction exists, it is added to the edge list of the network.

        Args:
            - cascades: A list of cascades, where each cascade is a sequence of nodes.

        Returns:
            - None
        """
        database = self.resources

        for cascade in cascades:
            interaction_in = database.loc[(database["source"] == cascade[0]) &
                                          (database["target"] == cascade[1])]
            if interaction_in.empty:
                print("Empty interaction for node ", cascade[0], " and ", cascade[1])
            else:
                self.add_edge(interaction_in)
        self.edges = self.edges.drop_duplicates().reset_index(drop=True)

        return

    def connect_nodes(self,
                      only_signed: bool = False,
                      consensus_only: bool = False
                      ) -> None:
        """
        Basic node connections. It adds all the interactions found in the omnipath database.
        Once an interaction is found it will be added to the list of edges.
        The only_signed flag makes sure that just signed interaction will be added to the network, while "consensus_only"
        makes sure that just signed interaction with consensus among references will be included.

        Args:
            - only_signed: A boolean flag indicating whether to only add signed interactions to the network.
            - consensus_only: A boolean flag indicating whether to only add signed interactions with consensus among
                            references to the network.

        Returns:
            - None
        """

        if len(self.nodes) == 1:
            print("Number of node insufficient to create connection")
            return

        def add_edge_if_not_empty_and_signed(node1, node2):
            """
            Helper function to add an edge to the network if the interaction is not empty and, if the `only_signed`
            flag is set, the interaction is signed.

            Args:
            - node1: The source node of the interaction.
            - node2: The target node of the interaction.

            Returns:
            None. The function modifies the network object in-place.
            """
            if node2 in self.__connect.find_all_neighbours(node1):
                interaction = self.resources.loc[(self.resources["source"] == node1) &
                                                 (self.resources["target"] == node2)]
                if not interaction.empty and (
                    not only_signed or check_sign(interaction, consensus_only) != "undefined"):
                    self.add_edge(interaction)

        for node1, node2 in combinations(self.nodes["Uniprot"], 2):
            add_edge_if_not_empty_and_signed(node1, node2)
            add_edge_if_not_empty_and_signed(node2, node1)
        return

    def is_connected(self) -> bool:
        """
        This function checks if the network is connected. It uses Depth-First Search (DFS) to traverse the network
        and check if all nodes are visited. If all nodes are visited, the network is connected.

        Args:
            - None

        Returns:
            - None

        """

        # Create a set to store visited nodes
        visited = set()

        # Function for Depth-First Search
        def dfs(node):
            visited.add(node)
            for neighbour in self.__connect.find_all_neighbours(node):
                if neighbour not in visited:
                    dfs(neighbour)

        # Start DFS from the first node
        dfs(self.nodes.iloc[0]['Uniprot'])

        # Check if all nodes are visited
        return set(self.nodes['Uniprot']) == visited

    def remove_undefined_interactions(self):
        """
        This function removes all undefined interactions from the network.

        Args:
            - None

        Returns:
            - None
        """
        self.edges = self.edges[self.edges['Effect'] != 'undefined']

    def __filter_unsigned_paths(self,
                                paths: list[tuple],
                                consensus: bool
                                ) -> list[tuple]:
        """
        This function filters out unsigned paths from the provided list of paths. An unsigned path is a path where at
        least one interaction does not have a defined sign (stimulation or inhibition). The function checks each
        interaction in each path, and if the interaction is unsigned, the path is not included in the output list.

        Args:
            - paths: A list of paths, where each path is a sequence of nodes.
            - consensus: A boolean indicating whether to check for consensus among references when determining the sign of an interaction.

        Returns:
            - A list[tuple] of paths where all interactions in each path are signed.
        """

        interactions = self.resources
        filtered_paths = []
        for path in paths:
            is_full_signed = True
            for i in range(0, len(path)):
                if i == len(path) - 1:
                    break
                interaction = interactions.loc[(interactions["source"] == path[i]) &
                                               (interactions["target"] == path[i + 1])]
                if not interaction.empty and check_sign(interaction, consensus) == "undefined":
                    is_full_signed = False
                    break
            if is_full_signed:
                filtered_paths.append(path)

        return filtered_paths

    def connect_subgroup(self,
                         group: (str | pd.DataFrame | list[str]),
                         maxlen: int = 1,
                         only_signed: bool = False,
                         consensus: bool = False
                         ) -> None:
        """
        This function is used to connect all the nodes in a particular subgroup. It iterates over all pairs of nodes
        in the subgroup and finds paths between them in the resources' database. If a path is found, it is added to
        the edge list of the network. The function also filters out unsigned paths if the `only_signed` flag is set
        to True.

        Args:
            - group: A list of nodes representing the subgroup to connect. Nodes can be represented as strings,
                        pandas DataFrame, or list of strings.
            - maxlen: The maximum length of the paths to be searched for in the resources' database. Default is 1.
            - only_signed: A boolean flag indicating whether to only add signed interactions to the network. Default is
                            False.
            - consensus: A boolean flag indicating whether to only add signed interactions with consensus among
                            references to the network. Default is False.

        Returns:
            - None
        """

        if not check_gene_list_format(group):
            uniprot_gene_list = group
        else:
            uniprot_gene_list = [mapping_node_identifier(i)[2] for i in group]
        if len(uniprot_gene_list) == 1:
            print("Number of node insufficient to create connection")
        else:
            for node1, node2 in combinations(uniprot_gene_list, 2):
                i = 0
                paths_in = []
                paths_out = []
                while i <= maxlen:
                    if not paths_out:
                        paths_out = self.__connect.find_paths(node1, node2, maxlen=i)
                        if only_signed:
                            paths_out = self.__filter_unsigned_paths(paths_out, consensus)
                    if not paths_in:
                        paths_in = self.__connect.find_paths(node2, node1, maxlen=i)
                        if only_signed:
                            paths_in = self.__filter_unsigned_paths(paths_in, consensus)
                    if not paths_in or not paths_out and i <= maxlen:
                        i += 1
                    if (paths_in or paths_out) and i > maxlen or (paths_in and paths_out):
                        paths = paths_out + paths_in
                        self.__add_paths_to_edge_list(paths)
                        break
        return

    def dfs_algorithm(self,
                      node1: str,
                      node2: str,
                      maxlen: int,
                      only_signed: bool,
                      consensus: bool,
                      connect_with_bias: bool
                      ) -> None:

        """

        This function uses the Depth-First Search (DFS) algorithm to find paths between two nodes in the network. It
        starts from the target node and searches for paths of increasing length until it finds a path of length
        maxlen. If the `only_signed` flag is set to True, it filters out unsigned paths. If the `connect_with_bias`
        flag is set to True, it connects the nodes when first introduced. Args: node1: node2: maxlen: only_signed:
        consensus: connect_with_bias:

        Args:
            - node1: A string representing the source node.
            - node2: A string representing the target node.
            - maxlen: An integer representing the maximum length of the paths to be searched for. Default is 2.
            - only_signed: A boolean flag indicating whether to filter unsigned paths. Default is False.
            - consensus: A boolean flag indicating whether to check for consensus among references. Default is False.
            - connect_with_bias: A boolean flag indicating whether to connect the nodes when first introduced.
                                Default is True.

        Returns:
            - None
        """

        i = 1
        min_len = 1
        while i <= maxlen:
            paths = self.__connect.find_paths(start=node1, end=node2, maxlen=i, minlen=min_len)
            if only_signed:
                paths = self.__filter_unsigned_paths(paths, consensus)
            if paths:
                self.__add_paths_to_edge_list(paths)
                if connect_with_bias:
                    self.connect_nodes(only_signed, consensus)
                    self.edges = self.edges.drop_duplicates().reset_index(drop=True)
                break
            else:
                i += 1
                min_len += 1

    def bfs_algorithm(self,
                      node1: str,
                      node2: str,
                      maxlen: int,
                      only_signed: bool,
                      consensus: bool,
                      connect_with_bias: bool
                      ) -> None:

        """

        This function uses the Breadth-First Search (BFS) algorithm to find paths between two nodes in the network.
        It starts from the target node and searches for paths of increasing length until it finds a path of length
        maxlen. If the `only_signed` flag is set to True, it filters out unsigned paths. If the `connect_with_bias`
        flag is set to True, it connects the nodes when first introduced. Args: node1: node2: only_signed: consensus:
        connect_with_bias:

        Args:
            - node1: A string representing the source node.
            - node2: A string representing the target node.
            - maxlen: An integer representing the maximum length of the paths to be searched for. Default is 2.
            - only_signed: A boolean flag indicating whether to filter unsigned paths. Default is False.
            - consensus: A boolean flag indicating whether to check for consensus among references. Default is False.
            - connect_with_bias: A boolean flag indicating whether to connect the nodes when first introduced.
                                Default is True.

        Returns:
            - None

        """

        paths = self.__connect.bfs(start=node1, end=node2)
        if only_signed:
            paths = self.__filter_unsigned_paths(paths, consensus)
        if paths:
            self.__add_paths_to_edge_list(paths)
            if connect_with_bias:
                self.connect_nodes(only_signed, consensus)
                self.edges = self.edges.drop_duplicates().reset_index(drop=True)

    def complete_connection(self,
                            maxlen: int = 2,
                            algorithm: Literal['bfs', 'dfs'] = 'dfs',
                            minimal: bool = True,
                            only_signed: bool = False,
                            consensus: bool = False,
                            connect_with_bias: bool = False
                            ) -> None:
        """
        This function attempts to connect all nodes of a network object using one of the methods presented in the
        Connection object. This is a core characteristic of this package and the user should have the possibility to
        choose different methods to enrich its Network object.

        Args:
            - maxlen: The maximum length of the paths to be searched for. Default is 2.
            - algorithm: The search algorithm to be used. It can be 'bfs' (Breadth-First Search) or 'dfs'
                (Depth-First Search).
            - minimal: A boolean flag indicating whether to reset the object connect_network, updating the possible list
                of paths. Default is True.
            - only_signed: A boolean flag indicating whether to filter unsigned paths. Default is False.
            - consensus: A boolean flag indicating whether to check for consensus among references. Default is False.
            - connect_with_bias: A boolean flag indicating whether to connect nodes when first
                introduced. Default is True.

        Returns:
            - None
        """

        # Copy the nodes
        nodes = self.nodes.copy()

        # Create a Connections object for the edges
        connect_network = Connections(self.edges)

        # Iterate through all combinations of nodes
        for node1, node2 in combinations(nodes["Uniprot"], 2):
            if not self.check_node(node1) or not self.check_node(node2):
                print(
                    "Error: node %s is not present in the resources database" % node1 if not self.check_node(
                        node1) else "Error: node %s is not present in the resources database" % node2)
                continue
            i = 0
            # Reset the object connect_network, updating the possible list of paths if minimal is True
            if minimal:
                connect_network = Connections(self.edges)

            # As first step, make sure that there is at least one path between two nodes in the network
            paths_in = connect_network.bfs(start=node2, end=node1)
            paths_out = connect_network.bfs(start=node1, end=node2)

            if not paths_in:
                self.__algorithms[algorithm](node1=node2, node2=node1, maxlen=maxlen, only_signed=only_signed, consensus=consensus, connect_with_bias=connect_with_bias)
            if not paths_out:
                self.__algorithms[algorithm](node1=node1, node2=node2, maxlen=maxlen, only_signed=only_signed, consensus=consensus, connect_with_bias=connect_with_bias)

        # If connect_with_bias is False, connect nodes after all paths have been found
        if not connect_with_bias:
            self.connect_nodes(only_signed, consensus)
            self.edges = self.edges.drop_duplicates().reset_index(drop=True)
        return

    def connect_component(self,
                          comp_A: (str | list[str]),
                          comp_B: (str | list[str]),
                          maxlen: int = 2,
                          mode: Literal['OUT', 'IN', 'ALL'] = 'OUT',
                          only_signed: bool = False,
                          consensus: bool = False
                          ) -> None:
        """
        This function attempts to connect subcomponents of a network object using one of the methods presented in the
        Connection object. This is a core characteristic of this package and the user should have the possibility to
        choose different methods to enrich its Network object.

        Args:
            - comp_A: A string or list of strings representing the first component to connect.
            - comp_B: A string or list of strings representing the second component to connect.
            - maxlen: The maximum length of the paths to be searched for. Default is 2.
            - mode: The search mode, which can be 'OUT', 'IN', or 'ALL'. Default is 'OUT'.
            - only_signed: A boolean flag indicating whether to filter unsigned paths. Default is False.
            - consensus: A boolean flag indicating whether to check for consensus among references. Default is False.

        Returns:
            - None

        """

        # Determine the search mode and find paths accordingly
        if mode == "IN":
            paths_in = self.__connect.find_paths(comp_B, comp_A, maxlen=maxlen)
            paths = paths_in
        elif mode == "OUT":
            paths_out = self.__connect.find_paths(comp_A, comp_B, maxlen=maxlen)
            paths = paths_out
        elif mode == "ALL":
            paths_out = self.__connect.find_paths(comp_A, comp_B, maxlen=maxlen)
            paths_in = self.__connect.find_paths(comp_B, comp_A, maxlen=maxlen)
            paths = paths_out + paths_in
        else:
            print("The only accepted modes are IN, OUT or ALL, please check the syntax")
            return

        # Filter unsigned paths if the only_signed flag is set
        if only_signed:
            paths = self.__filter_unsigned_paths(paths, consensus)

        # Add the paths to the edge list
        self.__add_paths_to_edge_list(paths)

        # Create sets of nodes for each component and the entire network
        all_nodes = set(self.nodes['Uniprot'].values)
        set_a = set(comp_A)
        set_b = set(comp_B)

        # Find the nodes that are not in either component
        set_c = all_nodes.difference(set_a)
        set_c = set_c.difference(set_b)
        set_c = list(set_c)

        # If there are nodes not in either component, connect them as a subgroup
        if len(set_c) > 0:
            self.connect_subgroup(set_c, only_signed=only_signed, maxlen=maxlen)

        return

    def connect_to_upstream_nodes(self,
                                  nodes_to_connect: Optional[List[str]] = None,
                                  depth: int = 1,
                                  rank: int = 1,
                                  only_signed: bool = True,
                                  consensus: bool = False
                                  ) -> None:
        """
        This function connects the provided nodes to their upstream nodes in the network.

        Args:
            - nodes_to_connect: A list of nodes to connect. If not provided, all nodes in the network are considered.
            - depth: The depth of the search for upstream nodes.
            - rank: The rank of the search for upstream nodes.
            - only_signed: A boolean flag indicating whether to filter unsigned paths. Default is True.
            - consensus: A boolean flag indicating whether to check for consensus among references. Default is False.

        Returns:
            - None
        """
        try:
            if nodes_to_connect is None:
                nodes_to_connect = self.nodes["Uniprot"].tolist()

            cascades = self.__connect.find_upstream_cascades(nodes_to_connect, depth, rank)

            if only_signed:
                cascades = self.__filter_unsigned_paths(cascades, consensus)
            self.__add_cascade_to_edge_list(cascades)
            self.edges.drop_duplicates().reset_index(drop=True)
        except Exception as e:
            print(f"An error occurred while connecting to upstream nodes: {e}")
        return

    def convert_edgelist_into_genesymbol(self) -> pd.DataFrame:
        """
        This function generates a new edges dataframe with the source and target identifiers translated (if possible)
        in Genesymbol format.

        Args:
             - None

        Returns:
            - A pandas DataFrame containing the edges with the source and target identifiers translated into Genesymbol
                format.
        """

        def convert_identifier(x):
            identifiers = mapping_node_identifier(x)
            return identifiers[0] or identifiers[1]

        gs_edges = self.edges.copy()

        gs_edges["source"] = gs_edges["source"].apply(convert_identifier)
        gs_edges["target"] = gs_edges["target"].apply(convert_identifier)

        return gs_edges

    def connect_genes_to_phenotype(self,
                                   phenotype: str = None,
                                   id_accession: str = None,
                                   sub_genes: list[str] = None,
                                   maxlen: int = 2,
                                   only_signed: bool = False,
                                   compress: bool = False
                                   ) -> None:
        """
        This function connects genes to a phenotype based on the provided Args. It retrieves phenotype markers,
        identifies unique Uniprot genes, and connects them to the network. It also has the option to compress the
        network by substituting specified genes with the phenotype name.

        Args:
            - phenotype: The phenotype to connect to. If not provided, it will be retrieved using the id_accession.
            - id_accession: The accession id of the phenotype. If not provided, the phenotype parameter must be given.
            - sub_genes: A list of genes to be considered for connection. If not provided, all nodes in the network are
                considered.
            - maxlen: The maximum length of the paths to be searched for.
            - only_signed: A boolean flag to indicate whether to filter unsigned paths.
            - compress: A boolean flag to indicate whether to substitute the specified genes with the phenotype name.

        Returns:
            - None
        """
        # Initialize lists for Uniprot and genesymbol genes
        uniprot_gene_list = []
        genesymbols_genes = []

        # Retrieve phenotype markers
        phenotype_genes = self.__ontology.get_markers(phenotype=phenotype, id_accession=id_accession)
        if not phenotype_genes:
            print("Something went wrong while getting the markers for: ", phenotype, " and ", id_accession)
            print("Check URL and try again")
            return
        # Convert phenotype genes to Uniprot identifiers
        uniprot_genes = [mapping_node_identifier(i)[2] for i in phenotype_genes]
        # If sub_genes are provided, check their format and convert to Uniprot or genesymbol as needed
        if sub_genes:
            if check_gene_list_format(sub_genes):
                uniprot_gene_list = sub_genes
                genesymbols_genes = [mapping_node_identifier(i)[2] for i in sub_genes]
            else:
                uniprot_gene_list = [mapping_node_identifier(i)[2] for i in sub_genes]
                genesymbols_genes = sub_genes

        print("Starting connecting network's nodes to: ", phenotype_genes)
        # Identify unique Uniprot genes not already in the network
        unique_uniprot = set(uniprot_genes) - set(uniprot_gene_list if uniprot_gene_list else self.nodes["Uniprot"])
        # Identify unique genesymbols not already in the network
        unique_genesymbol = set(phenotype_genes) - set(
            genesymbols_genes if genesymbols_genes else self.nodes["Genesymbol"])
        # Connect the network's nodes to the unique Uniprot genes
        self.connect_component(uniprot_gene_list if uniprot_gene_list else self.nodes["Uniprot"].tolist(),
                               list(unique_uniprot),
                               mode="OUT",
                               maxlen=maxlen, only_signed=only_signed)

        # If compress is True, substitute specified genes with the phenotype name
        if compress:
            phenotype = phenotype or self.__ontology.accession_to_phenotype_dict[id_accession]
            phenotype_modified = phenotype.replace(" ", "_")

            # Substitute the specified genes with the phenotype name in the nodes dataframe
            self.nodes['Uniprot'] = self.nodes['Uniprot'].apply(
                lambda x: phenotype_modified if x in unique_uniprot else x)
            self.nodes['Genesymbol'] = self.nodes['Genesymbol'].apply(
                lambda x: phenotype_modified if x in unique_genesymbol else x)

            # Substitute the specified genes with the phenotype name in the edges dataframe
            for column in ['source', 'target']:
                self.edges[column] = self.edges[column].apply(
                    lambda x: phenotype_modified if x in unique_uniprot else x)

            # Group by source and target, and aggregate with the custom function for each column
            self.edges = self.edges.groupby(['source', 'target']).agg({
                'Type': join_unique,  # Aggregate types with the custom function
                'Effect': determine_most_frequent_effect,  # Use the new function to determine the most frequent effect
                'References': join_unique  # Aggregate references with the custom function
            }).reset_index()

            # Identify common genes between Uniprot genes and the network's nodes
            common_genes = set(uniprot_genes).intersection(
                set(uniprot_gene_list if uniprot_gene_list else self.nodes["Uniprot"]))
            # For each common gene, add a new edge connecting the gene to the phenotype
            for gene in common_genes:
                new_edge = pd.DataFrame({"source": [gene], "target": [phenotype_modified], "Effect": ["stimulation"],
                                         "References": ["Gene Ontology"]})
                self.edges = pd.concat([self.edges, new_edge], ignore_index=True)
        return

    def connect_network_radially(self,
                                 max_len: int = 1,
                                 direction: Literal['OUT', 'IN', None] = None,
                                 loops: bool = False,
                                 consensus: bool = False,
                                 only_signed: bool = True
                                 ) -> None:

        """
        This function connects all nodes of a network object in a radial manner. It iteratively connects upstream and
        downstream nodes of the initial nodes. The function also removes any nodes that do not have a source in the edge
        dataframe and are not in the initial nodes.

        Args:
            - max_len: The maximum length of the paths to be searched for. Default is 1.
            - direction: The direction of the search. It can be 'OUT', 'IN', or None. Default is None.
            - loops: A boolean flag indicating whether to allow loops in the network. Default is False.
            - consensus: A boolean flag indicating whether to check for consensus among references. Default is False.
            - only_signed: A boolean flag indicating whether to filter unsigned paths. Default is True.

        Returns:
            - None

        """

        initial_nodes = self.initial_nodes
        initial_nodes_set = set([mapping_node_identifier(i)[2] for i in initial_nodes])

        i = 0
        source_nodes = initial_nodes_set
        target_nodes = initial_nodes_set

        while i < max_len:
            new_nodes = []
            if direction == 'OUT' or direction is None:
                for source in source_nodes:
                    target_neighs = self.__connect.find_target_neighbours(source)
                    if source in target_neighs and not loops:
                        target_neighs.remove(source)
                    target_paths = [(source, node) for node in target_neighs]
                    if only_signed:
                        target_paths = self.__filter_unsigned_paths(target_paths, consensus)
                    self.__add_paths_to_edge_list(target_paths)
                    target_neighs_filtered = [path[1] for path in target_paths]
                    target_neighs_filtered = [node for node in target_neighs_filtered if node not in initial_nodes_set]
                    new_nodes.extend(target_neighs_filtered)
                source_nodes = new_nodes

            new_nodes = []
            if direction == 'IN' or direction is None:
                for target in target_nodes:
                    source_neighs = self.__connect.find_source_neighbours(target)
                    if target in source_neighs and not loops:
                        source_neighs.remove(target)
                    source_paths = [(node, target) for node in source_neighs]
                    if only_signed:
                        source_paths = self.__filter_unsigned_paths(source_paths, consensus)
                    self.__add_paths_to_edge_list(source_paths)
                    source_neighs_filtered = [path[0] for path in source_paths]
                    source_neighs_filtered = [node for node in source_neighs_filtered if node not in initial_nodes_set]
                    new_nodes.extend(source_neighs_filtered)
                target_nodes = new_nodes
            i += 1

        # remove all nodes that have no source or that have no target and are not in the initial nodes
        target_nodes = set(self.edges["target"].unique())
        source_nodes = set(self.edges["source"].unique())

        disconnected_nodes = self.nodes[
            ~self.nodes["Uniprot"].isin(initial_nodes_set) & (
                ~self.nodes["Uniprot"].isin(target_nodes) | ~self.nodes["Uniprot"].isin(source_nodes))]

        # Loop until there are no more disconnected nodes
        while not disconnected_nodes.empty:
            # Collect nodes to remove
            nodes_to_remove = disconnected_nodes["Uniprot"].tolist()

            for node in nodes_to_remove:
                self.remove_node(node)

            # Recalculate disconnected nodes
            target_nodes = set(self.edges["target"].unique())
            source_nodes = set(self.edges["source"].unique())

            disconnected_nodes = self.nodes[
                ~self.nodes["Uniprot"].isin(initial_nodes_set) & (
                    ~self.nodes["Uniprot"].isin(target_nodes) | ~self.nodes["Uniprot"].isin(source_nodes))]

        return

    def connect_as_atopo(self,
                         strategy: Literal['radial', 'complete', None] = None,
                         max_len: int = 1,
                         loops: bool = False,
                         outputs=None,
                         only_signed: bool = True,
                         consensus: bool = False
                         ) -> None:
        """
        This method attempts to connect all nodes of a network object in a topological manner. It iteratively
        connects upstream nodes and checks if the network is connected. If not, it increases the search depth and
        repeats the process. It also removes any nodes that do not have a source in the edge dataframe and are not in
        the output nodes.

        Args:
            - strategy: The strategy to use to connect the network. It can be 'radial' or 'complete'. Default is None.
            - max_len: The maximum length of the paths to be searched for. Default is 1.
            - loops: A boolean flag indicating whether to allow loops in the network. Default is False.
            - outputs: A list of output nodes to connect to. Default is None.
            - only_signed: A boolean flag indicating whether to filter unsigned paths. Default is True.
            - consensus: A boolean flag indicating whether to check for consensus among references. Default is False.

        Returns:
            - None
        """

        initial_nodes = [mapping_node_identifier(i)[2] for i in self.initial_nodes]
        initial_nodes_set = set(initial_nodes)

        # Chose the strategy to use to connect the network
        if strategy == 'radial':
            self.connect_network_radially(max_len, direction=None,
                                          loops=loops, consensus=consensus, only_signed=only_signed)
        elif strategy == 'complete':
            self.complete_connection(max_len, minimal=True, only_signed=only_signed,
                                     consensus=consensus,
                                     connect_with_bias=False)
        else:
            pass

        starting_nodes = set(self.nodes["Uniprot"].tolist())

        if outputs is None:
            return

        # Add output nodes to the network
        for node in outputs:
            self.add_node(node)

        # Convert output nodes to Uniprot identifiers
        outputs_uniprot = [mapping_node_identifier(i)[2] for i in outputs]

        # Initialize depth
        depth = 1

        # While the network is not connected, connect to upstream nodes and first increase the rank and then the depth
        while not is_connected(self):
            self.connect_to_upstream_nodes(outputs_uniprot, depth=depth, rank=len(outputs_uniprot),
                                           only_signed=only_signed, consensus=consensus)
            new_nodes = set(self.nodes["Uniprot"].tolist()) - starting_nodes
            new_nodes = new_nodes - set(outputs_uniprot)

            # Remove nodes that do not have a source in the edge dataframe
            for node in new_nodes:
                if node not in self.edges["target"].unique():
                    self.remove_node(node)
                # remove a node if it auto-regulates itself
                if not loops and any((self.edges['source'] == node) & (self.edges['target'] == node)):
                    self.remove_node(node)

            # If depth reaches 4, stop the process
            if depth == 4:
                print("Current depth is 4, stopping the process")
                break

            # Increase depth
            depth += 1

        # Remove duplicate edges
        self.edges.drop_duplicates().reset_index(drop=True)

        # Create a set of unique sources from the edges DataFrame
        target_nodes = set(self.edges["target"].unique())

        # Identify nodes in the network that are not sources in the edges and not in initial nodes
        disconnected_nodes = self.nodes[
            ~self.nodes["Uniprot"].isin(initial_nodes_set) & ~self.nodes["Uniprot"].isin(target_nodes)]

        # Loop until there are no more disconnected nodes
        while not disconnected_nodes.empty:
            # Collect nodes to remove
            nodes_to_remove = disconnected_nodes["Uniprot"].tolist()

            for node in nodes_to_remove:
                self.remove_node(node)

            # Recalculate disconnected nodes
            target_nodes = set(self.edges["target"].unique())
            disconnected_nodes = self.nodes[
                ~self.nodes["Uniprot"].isin(initial_nodes_set) & ~self.nodes["Uniprot"].isin(target_nodes)]

        return



from __future__ import annotations

from typing import Literal


EFFECT_TYPES = {
    "1": "stimulation",
    "activate": "stimulation",
    "stimulate": "stimulation",
    "phosphorylate": "stimulation",
    "stimulation": "stimulation",
    "->": "stimulation",
    "-|": "inhibition",
    "-1": "inhibition",
    "inhibit": "inhibition",
    "block": "inhibition",
    "inhibition": "inhibition",
    "form complex": "form_complex",
    "form_complex": "form_complex",
    "form-complex": "form_complex",
    "complex formation": "form_complex"
}


def effect(
        value: str | int
    ) -> Literal['stimulation', 'inhibition', 'form_complex', 'undefined']:
    """
    Determine the effect based on the interaction type.

    Args:
        interaction_type:
            A string representing the type of interaction.

    Return:
       A string representing the effect of the interaction.
       If the interaction type is not recognized, it returns "undefined".
    """

    return effect_types.get(str(value).lower(), "undefined")



import pandas as pd

from ... import data as _data


def huri(dataset: str = 'HI-union') -> pd.DataFrame:

    df = (
        _data.huri(dataset, translated = True).
        drop(['source', 'target'], axis = 1).
        rename(columns = lambda x: x.split('_')[0]).
        drop_duplicates()
    )

    return df




import urllib.parse

import pandas as pd
from networkcommons.data.omics import _common as _ncdata


def _retrieve(path: str, ftype: str = 'tsv') -> pd.DataFrame:

    url = urllib.parse.urljoin(_ncdata._baseurl(), path)

    return _ncdata._open(url, ftype = ftype, df = True)


def phosphosite_kinase_substrate():

    return _retrieve('phosphosite/kinase-substrate.tsv')


def phosphosite_regulatory_sites():

    return _retrieve('phosphosite/regulatory-sites.tsv')


def huri(dataset: str = 'HI-union', translated = True) -> pd.DataFrame:

    url = (
        'https://github.com/sysbio-curie/Medulloblastoma_project/'
        'raw/main/Huri_analysis/data/%s%s.csv'
    )

    url = url % (dataset, '_translated' if translated else '')

    df = _ncdata._open(url, df = True)

    return df



import pandas as pd
from typing import Literal

from ... import data as _data


def psp(
    organism: Literal["human", "mouse", "rat"],
    kinase_substrate: str | None = None,
    regulatory_sites: str | None = None,
) -> pd.DataFrame:
    """
    PhoshpositePlus (PSP) from local file.

    Parses the files to create sign interactions based on the
    effect of phosphorylation on protein activities and creates an interaction dataframe based on the Omnipath
    interaction format.

    Parameters:
        kinase_substrate (str): The path to the kinase interaction file.
        regulatory_sites (str): The path to the phosphorylation effect file.
        organism (str): The organism to filter interactions for.

    Returns:
        ks (pd.DataFrame): The processed interactions dataframe.
    """

    args = locals()

    def load(dset):

        nonlocal args
        value = args[dset]

        if isinstance(value, str):

            value = pd.read_table(value, sep='\t')

        elif value is None:

            value = getattr(_data, f'phosphosite_{dset}')()

        if not isinstance(value, pd.DataFrame):
            raise ValueError(f"Invalid type for '{dset}' ({type(value)}).")

        return value

    ks, rs = tuple(load(dset) for dset in ('kinase_substrate', 'regulatory_sites'))

    # Filter the kinase interaction dataframe to include only interactions from the specified organism
    ks = ks.loc[
        (ks['KIN_ORGANISM'] == organism) &
        (ks['SUB_ORGANISM'] == organism)
    ]

    # Concatenate the SUB_GENE and SUB_MOD_RSD columns to create the target column
    ks['target'] = ks['SUB_GENE'] + '_' + ks['SUB_MOD_RSD']

    # Create the ks dataframe with the source and target columns
    ks = pd.DataFrame({
        'source': ks['GENE'],
        'target': ks['target'],
        'is_directed': True,
        'consensus_direction': False,  # Assuming no data provided, set all to False
        'consensus_stimulation': False,  # Assuming no data provided, set all to False
        'consensus_inhibition': False,  # Assuming no data provided, set all to False
    })

    # Filter the phosphorylation effect dataframe to include only interactions from the specified organism
    rs = rs.loc[rs['ORGANISM'] == organism]

    # Keep only MOD_RSD entries with "-p" suffix and remove it
    rs = rs[rs['MOD_RSD'].str.endswith('-p')]
    rs['MOD_RSD'] = rs['MOD_RSD'].apply(lambda x: x[:-2])

    # Concatenate the GENE and MOD_RSD columns to create the Prot_site column
    rs['Prot_site'] = rs['GENE'] + '_' + rs[
        'MOD_RSD']

    # Initialize the is_stimulation and is_inhibition columns with 0
    ks['is_stimulation'] = 0
    ks['is_inhibition'] = 0

    # Function to update the is_stimulation and is_inhibition columns based on the ON_FUNCTION values
    def update_phospho_effect(row):
        """
        This function updates the is_stimulation and is_inhibition columns based on the ON_FUNCTION values.

        Parameters:
        row (pd.Series): The row to update.

        Returns:
        row (pd.Series): The updated row.
        """
        if isinstance(row['ON_FUNCTION'], str):  # Check if the value is a string
            activities = []
            # Split each value in the ON_FUNCTION column
            for function in row['ON_FUNCTION'].split(';'):
                # Check if the function contains the word "activity"
                if 'activity' in function:
                    activities.append(function.strip())  # Add the function to the list of activities
            # Initialize the stimulation and inhibition flags
            is_stimulation = 0
            is_inhibition = 0
            # Check each activity for "induced" or "inhibited"
            for activity in activities:
                if 'induced' in activity:
                    is_stimulation = 1
                elif 'inhibited' in activity:
                    is_inhibition = 1
            # Update the row with the flags
            row['is_stimulation'] = is_stimulation
            row['is_inhibition'] = is_inhibition

        return row

    # Update the ks dataframe based on the phosphorylation effect information
    ks = ks.merge(rs[['Prot_site', 'ON_FUNCTION']],
                  left_on='target', right_on='Prot_site', how='left')
    ks.fillna('', inplace=True)
    ks = ks.apply(update_phospho_effect, axis=1)
    ks.drop(columns=['Prot_site', 'ON_FUNCTION'], inplace=True)

    # If expand is True, expand each line to include a new interaction from phosphosite to respective protein
    expanded_rows = []

    for index, row in ks.iterrows():

        source, target = row['source'], row['target']
        target_protein = target.split('_')[0]
        expanded_rows.append({'source': target, 'target': target_protein, 'is_directed': True,
                              'is_stimulation': row['is_stimulation'], 'is_inhibition': row['is_inhibition'],
                              'consensus_direction': False, 'consensus_stimulation': False,
                              'consensus_inhibition': False, 'curation_effort': False, 'references': False,
                              'sources': False})

    ks = pd.concat([ks, pd.DataFrame(expanded_rows)], ignore_index=True)

    # Filter the ks dataframe based on the is_stimulation and is_inhibition columns
    ks = ks[
        (ks['is_stimulation'] != 0) | (ks['is_inhibition'] != 0)]

    ks = ks.drop_duplicates(subset=['source', 'target'])

    return ks



from typing import Union, List, Tuple
import pandas as pd
from collections import deque
import random


class Connections:
    """
    Class that stores many utility functions to enrich an object Network.
    Each utility functions should take as input the nodes dataframe, which is used as base for each algorithm, and a
    database from the inputs modules, which will be used to extend the initial network.
    """

    def __init__(self, database: pd.DataFrame):
        self.resources = database.copy()
        self.target_neighbours_map = self._preprocess_target_neighbours()
        self.source_neighbours_map = self._preprocess_source_neighbours()

    def _preprocess_target_neighbours(self) -> dict:
        """
        Preprocess the targets neighbours map for fast lookup.
        """
        df_sources = self.resources.set_index('source')
        return df_sources.groupby('source')['target'].apply(set).to_dict()

    def _preprocess_source_neighbours(self) -> dict:
        """
        Preprocess the sources neighbours map for fast lookup.
        """
        df_targets = self.resources.set_index('target')
        return df_targets.groupby('target')['source'].apply(set).to_dict()

    def find_target_neighbours(self, node: str) -> List[str]:
        """
        Optimized helper function that finds the neighbors of the target node.
        """
        return list(self.target_neighbours_map.get(node, []))

    def find_source_neighbours(self, node: str) -> List[str]:
        """
        Optimized helper function that finds the neighbors of the target node.
        """
        return list(self.source_neighbours_map.get(node, []))

    def find_all_neighbours(self, node: str) -> List[str]:
        """
        Optimized helper function that finds all neighbors (both source and target) of the target node.
        """
        target_neighs = self.find_target_neighbours(node)
        source_neighs = self.find_source_neighbours(node)
        return list(set(target_neighs + source_neighs))

    def bfs(self, start: str, end: str) -> List[Tuple[str, ...]]:
        """
        Find a path between two nodes using Breadth First Search. Useful to quickly check if two nodes are connected.

        Args:
            start: The start node.
            end: The end node.

        Returns:
            List of nodes representing the path from start to end. If no path exists, returns an empty list.
        """
        visited = set()
        queue = deque([(start, [start])])  # Store the path along with the node

        while queue:
            node, path = queue.popleft()
            if node == end:
                return [(path[i], path[i + 1]) for i in range(len(path) - 1)]

            if node not in visited:
                visited.add(node)
                neighbours = self.find_target_neighbours(node)
                random.shuffle(neighbours)  # Shuffle the neighbours to avoid bias
                queue.extend((neighbour, path + [neighbour]) for neighbour in neighbours if neighbour not in visited)

        return []

    def find_paths(self,
                   start: Union[str, pd.DataFrame, List[str]],
                   end: Union[str, pd.DataFrame, List[str], None] = None,
                   maxlen: int = 2,
                   minlen: int = 1,
                   loops: bool = False) -> List[Tuple[str, ...]]:
        """
        Find paths or motifs in a network.
        """

        def convert_to_string_list(start):
            if isinstance(start, str):
                return [start]
            elif isinstance(start, pd.DataFrame):
                return start['name_of_node'].tolist()
            elif isinstance(start, list) and all(isinstance(item, str) for item in start):
                return start
            else:
                raise ValueError("Invalid type for 'start' variable")

        def find_all_paths_aux(start, end, path, maxlen):
            path = path + [start]

            if len(path) >= minlen + 1 and (start == end or (end is None and not loops and len(path) == maxlen + 1) or (
                loops and path[0] == path[-1])):
                return [path]

            paths = []

            if len(path) <= maxlen:
                next_steps = self.find_target_neighbours(start)

                if not loops:
                    next_steps = list(set(next_steps) - set(path))

                for node in next_steps:
                    paths.extend(find_all_paths_aux(node, end, path, maxlen))

            return paths

        start_nodes = convert_to_string_list(start)
        end_nodes = convert_to_string_list(end) if end else [None]

        minlen = max(1, minlen)
        all_paths = []

        for s in start_nodes:
            for e in end_nodes:
                all_paths.extend(find_all_paths_aux(s, e, [], maxlen))

        return all_paths

    def find_upstream_cascades(self,
                               target_genes: List[str],
                               max_depth: int = 1,
                               selected_rank: int = 1) -> List[Tuple[str, str]]:
        """
        Find cascades of interactions in the network.
        Parameters:
        - target_genes: List of target genes to start the cascade.
        - max_depth: Maximum depth of the cascade.
        - selected_rank: Number of top regulators to select for each iteration.
        Returns:
        - interactions: List of interactions in the cascade.
        """

        def collect_for_depth(current_targets, current_depth):
            if current_depth > max_depth:
                return []

            mcs_regulators = find_minimal_covering_regulators(self.resources, current_targets, selected_rank)

            interactions = [(reg, target) for reg in mcs_regulators for target in self.target_neighbours_map.get(reg, []) if target in current_targets]  # this is not working

            if current_depth < max_depth:
                next_targets = list(mcs_regulators)
                interactions += collect_for_depth(next_targets, current_depth + 1)

            return interactions

        return collect_for_depth(target_genes, 1) # it returns nothing


def remove_keys_and_related_values(dictionary: dict, keys_to_remove: list[str]) -> dict:
    """
    Remove keys and related values from a dictionary.
    """
    values_to_remove = set()
    for key in keys_to_remove:
        values_to_remove.update(dictionary.get(key, set))
        # Identify all keys to be removed based on the collected values
    all_keys_to_remove = set(keys_to_remove)
    for key, values in dictionary.items():
        if not values.isdisjoint(values_to_remove):  # If there's any overlap in values
            all_keys_to_remove.add(key)

    # Remove all identified keys from the dictionary
    for key in all_keys_to_remove:
        dictionary.pop(key, None)

    return dictionary


def check_dataframe_coverage(df: pd.DataFrame, target_genes: List[str]) -> List[str]:
    """
    Checks the DataFrame for the presence of the specified target genes and returns
    an updated list with those genes that are covered.
    """
    present_targets = set(df['target'].unique())
    targets = set(target_genes)
    covered_targets = list(targets & present_targets)
    return covered_targets


def find_minimal_covering_regulators(df: pd.DataFrame, target_genes: list[str], rank: int = 1):
    """
    Find the minimal set of regulators that cover all target genes in the DataFrame.

    Parameters:
    - df: pandas DataFrame with at least 'source' and 'target' columns.
    - target_genes: List of target genes to cover.
    - rank: Number of top regulators to select for each iteration.

    Returns:
    - regulators: Set of regulators that cover all target genes.
    """
    # Update target_genes list based on DataFrame coverage
    covered_targets = check_dataframe_coverage(df, target_genes)

    if len(covered_targets) < len(target_genes):
        missing_targets = set(target_genes) - set(covered_targets)
        print(f"Warning: Some target genes are not present in the DataFrame and will be ignored: {missing_targets}")

    # Proceed to filter the DataFrame for covered target genes
    filtered_df = df[df['target'].isin(covered_targets)]
    regulators = set()
    targets_covered = set()

    # Find unique regulators for each gene
    regulator_to_targets = filtered_df.groupby('source')['target'].apply(set).to_dict()

    while targets_covered != set(covered_targets):

        previous_covered_count = len(targets_covered)  # Keep track of previously covered targets
        # Calculate the lengths of values for each regulator
        lengths = sorted({len(values) for values in regulator_to_targets.values()}, reverse=True)

        # Select the top_n lengths
        selected_lengths = lengths[:rank]

        # Find all keys that have values lengths in the selected_lengths
        keys_with_max_values = [key for key, values in regulator_to_targets.items() if
                                len(values) in selected_lengths]

        regulators.update(keys_with_max_values)

        for reg in keys_with_max_values:
            targets_covered.update(regulator_to_targets[reg])

        # Check if no new targets were covered in this iteration
        if len(targets_covered) == previous_covered_count:
            print("Warning: Unable to cover all targets with available regulators. Exiting...")
            break
        regulator_to_targets = remove_keys_and_related_values(regulator_to_targets, keys_with_max_values)

    return regulators



import omnipath as op
import pandas as pd
import logging
import pypath
from pypath.utils import mapping

from . import _misc as _misc

"""
Access to network databases.
"""


def omnipath_universe(**kwargs):
    """
    Access generic networks from OmniPath.
    """

    return op.interactions.PostTranslational.get(**kwargs)





from __future__ import annotations
import pandas as pd
from typing_extensions import Literal


def remove_keys_and_related_values(dictionary: dict, keys_to_remove: list[str]) -> dict:
    """
    Remove keys and related values from a dictionary.
    """
    values_to_remove = set()
    for key in keys_to_remove:
        values_to_remove.update(dictionary.get(key, set))
        # Identify all keys to be removed based on the collected values
    all_keys_to_remove = set(keys_to_remove)
    for key, values in dictionary.items():
        if not values.isdisjoint(values_to_remove):  # If there's any overlap in values
            all_keys_to_remove.add(key)

    # Remove all identified keys from the dictionary
    for key in all_keys_to_remove:
        dictionary.pop(key, None)

    return dictionary


def check_dataframe_coverage(df: pd.DataFrame, target_genes: list[str]) -> list[str]:
    """
    Checks the DataFrame for the presence of the specified target genes and returns
    an updated list with those genes that are covered.

    Parameters:
    - df: pandas DataFrame with at least 'target' column.
    - target_genes: Iterable of target genes to check for in the DataFrame.

    Returns:
    - covered_targets: List of target genes that are present in the DataFrame.
    """
    # Determine which target genes are present in the DataFrame
    present_targets = set(df['target'].unique())
    targets = set(target_genes)

    # Filter out target genes not present in the DataFrame
    covered_targets = list(targets & present_targets)

    return covered_targets


def find_minimal_covering_regulators(df: pd.DataFrame, target_genes: list[str], rank: int = 1):
    """
    Find the minimal set of regulators that cover all target genes in the DataFrame.

    Parameters:
    - df: pandas DataFrame with at least 'source' and 'target' columns.
    - target_genes: List of target genes to cover.
    - rank: Number of top regulators to select for each iteration.

    Returns:
    - regulators: Set of regulators that cover all target genes.
    """
    # Update target_genes list based on DataFrame coverage
    covered_targets = check_dataframe_coverage(df, target_genes)

    if len(covered_targets) < len(target_genes):
        missing_targets = set(target_genes) - set(covered_targets)
        print(f"Warning: Some target genes are not present in the DataFrame and will be ignored: {missing_targets}")

    # Proceed to filter the DataFrame for covered target genes
    filtered_df = df[df['target'].isin(covered_targets)]
    regulators = set()
    targets_covered = set()

    # Find unique regulators for each gene
    regulator_to_targets = filtered_df.groupby('source')['target'].apply(set).to_dict()

    while targets_covered != set(covered_targets):

        previous_covered_count = len(targets_covered)  # Keep track of previously covered targets
        # Calculate the lengths of values for each regulator
        lengths = sorted({len(values) for values in regulator_to_targets.values()}, reverse=True)

        # Select the top_n lengths
        selected_lengths = lengths[:rank]

        # Find all keys that have values lengths in the selected_lengths
        keys_with_max_values = [key for key, values in regulator_to_targets.items() if
                                len(values) in selected_lengths]

        regulators.update(keys_with_max_values)

        for reg in keys_with_max_values:
            targets_covered.update(regulator_to_targets[reg])

        # Check if no new targets were covered in this iteration
        if len(targets_covered) == previous_covered_count:
            print("Warning: Unable to cover all targets with available regulators. Exiting...")
            break
        regulator_to_targets = remove_keys_and_related_values(regulator_to_targets, keys_with_max_values)

    return regulators


class Connections:
    """
    Class that stores many utility functions to enrich an object Network.
    Each utility functions should take as input the nodes dataframe, which is used as base for each algorithm, and a
    database from the inputs modules, which will be used to extend the initial network.
    """

    def __init__(self, database):
        self.resources = database
        return

    def find_neighbours(self,
                        node: str) -> list[str]:
        """
        Optimized helper function that finds the neighbors of the target node.
        """

        db = self.resources

        neigh = set(db.loc[db["source"] == node]["target"])

        return list(neigh)

    def find_paths(self,
                   start: (
                       str | pd.DataFrame | list[str]
                   ),
                   end: (
                       str | pd.DataFrame | list[str] | None
                   ) = None,
                   maxlen: int = 2,
                   minlen: int = 1,
                   loops: bool = False,
                   ) -> list[tuple]:
        """
        Find paths or motifs in a network.
        Adapted from pypath function 'find_paths' in Network core class.
        In future this class will be extended to take into account many other parameters to select those paths that
        match certain criteria (type of interaction, effect, direction, data model, etc...)
        For now (version 0.1.0) it will act as skeleton to retrieve interactions from AllInteractions in Omnipath.
        """

        def convert_to_string_list(start):
            if isinstance(start, str):
                return [start]
            elif isinstance(start, pd.DataFrame):
                return start['name_of_node'].tolist()
            elif isinstance(start, list) and all(isinstance(item, str) for item in start):
                return start
            else:
                raise ValueError("Invalid type for 'start' variable")

        def find_all_paths_aux(start, end, path, maxlen):
            path = path + [start]

            if len(path) >= minlen + 1 and (start == end or (end is None and not loops and len(path) == maxlen + 1) or (
                loops and path[0] == path[-1])):
                return [path]

            paths = []

            if len(path) <= maxlen:
                next_steps = self.find_neighbours(start)

                if not loops:
                    next_steps = list(set(next_steps) - set(path))

                for node in next_steps:
                    paths.extend(find_all_paths_aux(node, end, path, maxlen))

            return paths

        start_nodes = convert_to_string_list(start)
        end_nodes = convert_to_string_list(end) if end else [None]

        minlen = max(1, minlen)
        all_paths = []

        for s in start_nodes:
            for e in end_nodes:
                all_paths.extend(find_all_paths_aux(s, e, [], maxlen))

        return all_paths

    def find_upstream_cascades(self,
                               target_genes: list[str],
                               max_depth: int = 1,
                               selected_rank: int = 1) -> list[tuple]:

        """
        Find cascades of interactions in the network.
        Parameters:
        - target_genes: List of target genes to start the cascade.
        - max_depth: Maximum depth of the cascade.
        - selected_rank: Number of top regulators to select for each iteration.
        Returns:
        - interactions: List of interactions in the cascade.
        """
        def collect_for_depth(current_targets, current_depth):
            # Base case: Exit if maximum depth is exceeded
            if current_depth > max_depth:
                return []

            # Find the minimal covering set of regulators for current targets
            mcs_regulators = find_minimal_covering_regulators(self.resources, current_targets, selected_rank)

            # Collect interactions (source, target) for these regulators
            interactions = [(reg, target) for reg in mcs_regulators for target in self.resources[self.resources['source'] == reg]['target'] if
                            target in current_targets]

            # If additional depth is required, treat the regulators just found as the next targets
            if current_depth < max_depth:
                next_targets = mcs_regulators  # The regulators themselves become the targets for the next depth
                interactions += collect_for_depth(next_targets, current_depth + 1)

            return interactions

        # Kick off the recursive process starting with the original genes as targets
        return collect_for_depth(target_genes, 1)



import pandas as pd
import os
import itertools
import re

class Exports:
    """
    This class implement many methods used to export the Network object in different format.
    In particular the exports format will be methods-oriented (MaBoSS, Ginsim, cobrexa and so on...).
    To start with, the user can export the Network in SIF and Bnet format.
    In the future many more versatile methods will be implemented (SBML) and annotations will be included for each
    interaction, including the DOI of the relative reference and annotations from each database
    """
    def __init__(self, network):
        net = network.copy()
        df_edges = net.convert_edgelist_into_genesymbol()
        self.nodes = net.nodes
        self.interactions = df_edges
        return

    def export_bnet(self, file_name="logic_model.bnet"):
        """
        Function to export the network in bnet format, creating multiple files for bimodal interactions.
        """
        # Checks for nodes and interactions data
        if not isinstance(self.nodes, pd.DataFrame) or self.nodes.empty:
            print("Error: Nodes data is missing or empty.")
            return
        if not isinstance(self.interactions, pd.DataFrame) or self.interactions.empty:
            print("Error: Interactions data is missing or empty.")
            return

        # Identify undefined interactions
        undefined_interactions = self.interactions.query("Effect == 'undefined'")
        if not undefined_interactions.empty:
            print(f"Warning: The network has {len(undefined_interactions)} UNDEFINED interaction(s).")
            print("Undefined interactions:")
            for index, row in undefined_interactions.iterrows():
                print(f"{row['source']} -> {row['target']}")
                print(f"Reference: {row['References']}")

        # Identify bimodal interactions
        bimodal_interactions = self.interactions.query("Effect == 'bimodal'")
        if not bimodal_interactions.empty:
            print(f"Warning: The network has {len(bimodal_interactions)} BIMODAL interaction(s).")
            print("Bimodal interactions:")
            for index, row in bimodal_interactions.iterrows():
                print(f"{row['source']} -> {row['target']}")
                print(f"Reference: {row['References']}")

        # Generate permutations for bimodal interactions
        bimodal_sources = bimodal_interactions['source'].tolist()
        bimodal_targets = bimodal_interactions['target'].tolist()
        permutations = list(itertools.product(['stimulation', 'inhibition'], repeat=len(bimodal_interactions)))

        # Create a directory for the BNet files
        os.makedirs(os.path.dirname(file_name), exist_ok=True)

        # Iterate through permutations and create a BNet file for each
        for i, perm in enumerate(permutations):
            # Create a copy of the interactions DataFrame
            interactions_copy = self.interactions.copy()

            # Update bimodal interactions based on the current permutation
            for j, (source, target) in enumerate(zip(bimodal_sources, bimodal_targets)):
                interactions_copy.loc[(interactions_copy['source'] == source) &
                                      (interactions_copy['target'] == target), 'Effect'] = perm[j]

            # Pre-filter stimulations, inhibitions, and exclude undefined effects
            stimulations = interactions_copy.query("Effect == 'stimulation'")
            inhibitions = interactions_copy.query("Effect == 'inhibition'")
            complex_formation = interactions_copy.query("Effect == 'form complex'")

            # Generate the file name for this permutation
            perm_file_name = f"{os.path.splitext(file_name)[0]}_{i + 1}.bnet"

            with open(perm_file_name, "w") as f:
                f.write("# model in BoolNet format\n")
                f.write("targets, factors\n")

                for entry in self.nodes.values:
                    node = entry[0]

                    # Replace special characters in node names
                    node = re.sub(r"[\/\-\s\#]", "_", node)

                    formula_on = [re.sub(r"[\/\-\s\#]", "_", src) for src in
                                  stimulations[stimulations["target"] == node]["source"].to_list()]
                    formula_off = [re.sub(r"[\/\-\s\#]", "_", src) for src in
                                   inhibitions[inhibitions["target"] == node]["source"].to_list()]
                    formula_complex = [re.sub(r"[\/\-\s\#]", "_", src) for src in
                                       complex_formation[complex_formation["target"] == node]["source"].to_list()]

                    # Constructing the formula
                    formula_parts = []
                    if formula_complex:
                        formula_parts.append(f"({' & '.join(formula_complex)})")
                    if formula_on:
                        formula_parts.append(f"({' | '.join(formula_on)})")
                    if formula_off:
                        formula_parts.append("!({})".format(" | ".join(formula_off)))

                    # Writing the node and its formula to the file
                    f.write(f"{node}, {' & '.join(formula_parts) if formula_parts else node}\n")

            print(f"Created BNet file: {perm_file_name}")

        print(f"Generated {len(permutations)} BNet files.")

    def export_sif(self, file_name="logic_model.sif"):
        """
        Function to export the network in SIF format
        """

        with open(file_name, 'w') as file:
            for index, row in self.interactions.iterrows():
                # Use the Effect column directly assuming it contains "activate" or "inhibit"
                interaction_type = row['Effect']

                if interaction_type == "form complex":
                    interaction_type = "form_complex"

                # Reference for the interaction
                interaction_reference = row['References']  # Adjust column name if necessary

                # Write a comment line with the interaction reference
                file.write(f"# Reference PMID: {interaction_reference}\n")

                # Write the formatted interaction to the .sif file
                file.write(f"{row['source']}\t{interaction_type}\t{row['target']}\n")

        return



import re
import requests
import pandas as pd
import omnipath as op


def fetch_nodes_from_url(url):

    """
    fetch the nodes in a list from the given geneontology url
    """
    print("Start requesting genes from Gene Ontology")
    print("Fetching from: ", url)
    response = requests.get(url)
    print("Done")
    if not response:
        print("Error fetching the genes, no entry found, please check the id accession code")
        return
    genes = response.text.split("\n")
    genes = genes[:-1]
    genes_unique = list(set(genes))
    return genes_unique


class Ontology:
    """
    class that stores some functionalities to connect phenotypes to nodes
    and to associate information for each node at tissue level

    """
    def __init__(self):
        self.gene_ontology_url = "https://golr-aux.geneontology.io/solr/select?defType=edismax&qt=standard&indent=on&wt=csv&rows=100000&start=0&fl=bioentity_label&facet=true&facet.mincount=1&facet.sort=count&json.nl=arrarr&facet.limit=25&hl=true&hl.simple.pre=%3Cem%20class=%22hilite%22%3E&hl.snippets=1000&csv.encapsulator=&csv.separator=%09&csv.header=false&csv.mv.separator=%7C&fq=document_category:%22annotation%22&fq=isa_partof_closure:%22GO:0007049%22&fq=taxon_subset_closure_label:%22Homo%20sapiens%22&fq=type:%22protein%22&fq=annotation_class_label:%22G1/S%20transition%20of%20mitotic%20cell%20cycle%22&facet.field=aspect&facet.field=taxon_subset_closure_label&facet.field=type&facet.field=evidence_subset_closure_label&facet.field=regulates_closure_label&facet.field=isa_partof_closure_label&facet.field=annotation_class_label&facet.field=qualifier&facet.field=annotation_extension_class_closure_label&facet.field=assigned_by&facet.field=panther_family_label&q=*:*"
        self.accession_to_phenotype_dict = {"GO:0010718": "positive regulation of epithelial to mesenchymal transition"}
        return

    def modify_url_ontology(self, new_go_code, new_description):
        """
        Modifies a given URL by replacing a GO code and a descriptive string
        in very specific locations identified by prefixes.

        Parameters:
        - url (str): The original URL to be modified.
        - new_go_code (str): The new GO code to insert into the URL.
        - new_description (str): The new descriptive string to insert into the URL.

        Returns:
        - str: The modified URL.
        """
        # Define the prefixes that identify where the replacements should occur,
        go_code_prefix = "isa_partof_closure:%22"
        description_prefix = "annotation_class_label:%22"

        # Patterns to find the exact locations for replacements
        go_code_pattern = re.escape(go_code_prefix) + r"GO:\d{7}"
        description_pattern = re.escape(description_prefix) + r".+?%22"

        # Perform the replacements
        url = re.sub(go_code_pattern, go_code_prefix + new_go_code, self.gene_ontology_url, 1)
        # URL-encode the new description
        new_description_encoded = re.sub(r" ", "%20", new_description)
        url = re.sub(description_pattern, description_prefix + new_description_encoded + "%22", url, 1)

        return url

    def get_markers(self,
                    phenotype: str = None,
                    id_accession: str = None):
        if phenotype and id_accession:
            self.accession_to_phenotype_dict[id_accession] = phenotype
            url = self.modify_url_ontology(id_accession, phenotype)
            genes = fetch_nodes_from_url(url)
            return genes
        elif phenotype and not id_accession:
            id_accession = {i for i in self.accession_to_phenotype_dict if self.accession_to_phenotype_dict[i]==phenotype}
            url = self.modify_url_ontology(id_accession, phenotype)
            genes = fetch_nodes_from_url(url)
            return genes
        elif id_accession and not phenotype:
            phenotype = self.accession_to_phenotype_dict[id_accession]
            url = self.modify_url_ontology(id_accession, phenotype)
            genes = fetch_nodes_from_url(url)
            return genes
        else:
            print("Invalid GO id accession or phenotype description:")
            print("GO if accession used = ", id_accession)
            print("Phenotype description used = ", phenotype)
            return

    def check_tissue_annotations(self, genes_df, tissue):
        """
        Check if tissue annotations for each gene symbol contain the specified tissue.

        Args:
        genes_df (DataFrame): DataFrame containing gene symbols.
        tissue (str): Tissue to check for in annotations.

        Returns:
        DataFrame: DataFrame with results indicating whether each gene symbol has tissue annotations containing the specified tissue.
        """

        results = []

        # Iterate over each gene symbol
        for genesymbol in genes_df['Genesymbol']:
            # Fetch annotations for the gene symbol
            annotations_df = op.requests.Annotations.get([genesymbol], force_full_download=True)

            # Filter annotations for tissue annotations
            tissue_annotations = annotations_df[annotations_df['label'] == 'tissue']

            # Test if any tissue annotation contains the specified tissue
            in_tissue = any(tissue.lower() in value.lower() for value in tissue_annotations['value'])

            # Append result to list
            results.append({'Genesymbol': genesymbol, 'in_tissue': in_tissue})

        # Create DataFrame from results
        results_df = pd.DataFrame(results)

        return results_df



import pandas as pd


def signor(path: str) -> pd.DataFrame:
    """
    SIGNOR database from TSV.

    Processes SIGNOR interactions from a local TSV file.

    Parameters:
        path (str):
            The path to the SIGNOR TSV.

    Returns:
        None
    """

    # Read the SIGNOR database file into a pandas DataFrame
    df = pd.read_table(path)

    def contains(field, substr):

        return substr in field

    # Filter out the rows where EFFECT is "form complex" or "unknown"
    df = df[df['EFFECT'] != "unknown"]

    # Transform the original dataframe into the desired format
    df = pd.DataFrame({
        'source': df['IDA'],
        'target': df['IDB'],
        'is_directed': df['DIRECT'],
        'is_stimulation': df['EFFECT'].apply(lambda x: contains(x, 'up-regulates')),
        'is_inhibition': df['EFFECT'].apply(lambda x: contains(x, 'down-regulates')),
        'form_complex': df['EFFECT'].apply(lambda x: contains(x, 'complex')),
        'consensus_direction': False,  # Assuming no data provided, set all to False
        'consensus_stimulation': False,  # Assuming no data provided, set all to False
        'consensus_inhibition': False,  # Assuming no data provided, set all to False
        'curation_effort': df['ANNOTATOR'],
        'references': df['PMID'],
        'sources': df['SIGNOR_ID'],
    })

    # Add the transformed DataFrame to the existing database
    df = _group_by_source_target(df)

    return df


def _group_by_source_target(df_ungrouped: pd.DataFrame) -> pd.DataFrame:

    grouped = df_ungrouped.groupby(['source', 'target'])

    def join_strings(series):
        return '; '.join(set(series.astype(str)))

    def majority_true(series):
        # Count True values and compare to half the length of the series
        return series.sum() >= (len(series) / 2)

    def process_interaction(df):
        new_df = pd.DataFrame({
            'source': df['source'],
            'target': df['target'],
            'is_directed': df['is_directed'],
            'is_stimulation': df['is_stimulation'],
            'is_inhibition': df['is_inhibition'],
            'form_complex': df['form_complex'],
            'consensus_direction': True,
            'consensus_stimulation': (
                df['is_stimulation'].sum() >
                df['is_inhibition'].sum()
            ),
            'consensus_inhibition':  (
                df['is_inhibition'].sum() >
                df['is_stimulation'].sum()
            ),
            'curation_effort': df['curation_effort'],
            'references': df['references'],
            'sources': df['sources']
        })
        return new_df

    new_groups = []

    for (source, target), group_data in grouped:
        if len(group_data) > 1:
            aggregate_functions = {
                "is_stimulation": majority_true,
                "is_inhibition": majority_true,
                "form_complex": majority_true,
                "curation_effort": join_strings,
                "references": join_strings,
                "sources": join_strings
            }
            new_df = group_data.groupby(
                ["source", "target", "is_directed", "consensus_direction", "consensus_stimulation",
                 "consensus_inhibition"]).aggregate(aggregate_functions).reset_index()
            new_groups.append(process_interaction(new_df))
        else:
            new_groups.append(group_data)

    # Concatenate all groups back into a single DataFrame
    df_unique = pd.concat(new_groups, ignore_index=True)

    return df_unique




import unittest
import pandas as pd
import os
import tempfile
from unittest.mock import patch, MagicMock
from unipressed.id_mapping.types import From, To
from unipressed import IdMappingClient


class TestIDTranslator(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.input_file = os.path.join(self.temp_dir, 'test_input.csv')
        self.output_file = os.path.join(self.temp_dir, 'test_output.csv')

        # Create a sample input file
        df = pd.DataFrame({
            'source': ['ENSG00000139618', 'ENSG00000141510', 'ENSG00000157764'],
            'target': ['ENSG00000155657', 'ENSG00000121879', 'ENSG00000171862']
        })
        df.to_csv(self.input_file, index=False)

        self.translator = IDTranslator(
            input_file=self.input_file,
            output_file=self.output_file,
            source_type=From.ENSEMBL,
            dest_type=To.UNIPROTKB,
            pickle_dir=self.temp_dir
        )

    def tearDown(self):
        # Clean up temporary files
        for file in os.listdir(self.temp_dir):
            os.remove(os.path.join(self.temp_dir, file))
        os.rmdir(self.temp_dir)

    def test_load_data(self):
        self.translator.load_data()
        self.assertIsNotNone(self.translator.df)
        self.assertEqual(len(self.translator.df), 3)
        self.assertEqual(len(self.translator.unique_ids), 6)

    @patch('id_translator.IdMappingClient.submit')
    def test_get_id(self, mock_submit):
        mock_submit.return_value = MagicMock()
        mock_submit.return_value.each_result.return_value = [{'to': 'P12345'}]

        result = self.translator.get_id(('ENSG00000139618', From.ENSEMBL, To.UNIPROTKB, 3))
        self.assertEqual(result, ('ENSG00000139618', 'P12345'))

    @patch('id_translator.IdMappingClient.submit')
    def test_translate(self, mock_submit):
        mock_submit.return_value = MagicMock()
        mock_submit.return_value.each_result.return_value = [{'to': 'P12345'}]

        self.translator.translate()
        self.assertTrue(os.path.exists(self.output_file))

        df = pd.read_csv(self.output_file)
        self.assertEqual(len(df), 3)
        self.assertTrue('source_UniProtKB' in df.columns)
        self.assertTrue('target_UniProtKB' in df.columns)

    def test_analyze_results(self):
        # Create a sample translated DataFrame
        df = pd.DataFrame({
            'source': ['ENSG1', 'ENSG2', 'ENSG3'],
            'target': ['ENSG4', 'ENSG5', 'ENSG6'],
            'source_UniProtKB': ['P1', 'P2', None],
            'target_UniProtKB': ['P4', None, 'P6']
        })
        self.translator.df = df

        with self.assertLogs(level='INFO') as cm:
            self.translator.analyze_results()

        self.assertIn("Translation success rate: 66.67%", cm.output[0])
        self.assertIn("Untranslated entries: 2 out of 6", cm.output[1])

    def test_remove_untranslated_entries(self):
        # Create a sample translated DataFrame
        df = pd.DataFrame({
            'source': ['ENSG1', 'ENSG2', 'ENSG3'],
            'target': ['ENSG4', 'ENSG5', 'ENSG6'],
            'source_UniProtKB': ['P1', 'P2', None],
            'target_UniProtKB': ['P4', None, 'P6']
        })
        self.translator.df = df

        cleaned_output = os.path.join(self.temp_dir, 'cleaned_output.csv')
        self.translator.remove_untranslated_entries(cleaned_output)

        cleaned_df = pd.read_csv(cleaned_output)
        self.assertEqual(len(cleaned_df), 1)  # Only one row should remain
        self.assertEqual(cleaned_df.iloc[0]['source'], 'ENSG1')
        self.assertEqual(cleaned_df.iloc[0]['target'], 'ENSG4')


if __name__ == '__main__':
    unittest.main()



import os
import pandas as pd
import pickle
from multiprocessing import Pool, cpu_count
import time
from tqdm.auto import tqdm
import logging
from typing import List, Dict, Union
from unipressed import IdMappingClient
import itertools
import random
from IPython.display import display, HTML

class IDTranslator:
    def __init__(self, input_file: str, output_file: str, source_type: str, dest_type: str,
                 pickle_dir: str = './pickles', batch_size: int = 100, processes: int = None,
                 columns: List[str] = None, max_retries: int = 3, clear_progress: bool = False,
                 input_columns: Union[List[str], Dict[str, str]] = None, has_header: bool = True,
                 multiple_mapping_strategy: str = 'expand'):
        self.input_file = input_file
        self.output_file = output_file
        self.source_type = source_type
        self.dest_type = dest_type
        self.pickle_dir = pickle_dir
        self.batch_size = batch_size
        self.processes = processes or max(1, cpu_count() - 1)
        self.columns = columns or ['source', 'target']
        self.max_retries = max_retries
        self.input_columns = input_columns
        self.has_header = has_header
        self.df = None
        self.id_mapping = {}
        self.unique_ids = set()
        self.multiple_mapping_strategy = multiple_mapping_strategy

        # Set up logging
        self.logger = logging.getLogger(f"{self.__class__.__name__}_{id(self)}")
        self.logger.setLevel(logging.INFO)

        # Create formatter
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        # Create file handler
        fh = logging.FileHandler('id_translator.log')
        fh.setLevel(logging.INFO)
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)

        # Create stream handler for notebook display
        class NotebookHandler(logging.Handler):
            def emit(self, record):
                msg = self.format(record)
                display(HTML(f"<pre>{msg}</pre>"))

        nh = NotebookHandler()
        nh.setLevel(logging.INFO)
        nh.setFormatter(formatter)
        self.logger.addHandler(nh)

        logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

        os.makedirs(self.pickle_dir, exist_ok=True)

        if clear_progress:
            self.clear_all_progress()

    def load_data(self):
        file_extension = os.path.splitext(self.input_file)[1].lower()

        if file_extension == '.csv':
            self.df = pd.read_csv(self.input_file, header=0 if self.has_header else None)
        elif file_extension in ['.xls', '.xlsx']:
            self.df = pd.read_excel(self.input_file, header=0 if self.has_header else None)
        elif file_extension == '.tsv':
            self.df = pd.read_csv(self.input_file, sep='\t', header=0 if self.has_header else None)
        else:
            raise ValueError(f"Unsupported file format: {file_extension}")

        if not self.has_header:
            self.df.columns = [f'col_{i}' for i in range(len(self.df.columns))]

        self.process_columns()

        self.unique_ids = set()
        for column in self.columns:
            self.unique_ids.update(self.df[column].unique())
        self.logger.info(f"Loaded {len(self.df)} rows with {len(self.unique_ids)} unique IDs")

    def process_columns(self):
        if isinstance(self.input_columns, dict):
            self.df.rename(columns=self.input_columns, inplace=True)
            self.columns = list(self.input_columns.values())
        elif isinstance(self.input_columns, list):
            if len(self.input_columns) > len(self.df.columns):
                raise ValueError("More column names provided than exist in the file")
            self.df.columns = self.input_columns + list(self.df.columns[len(self.input_columns):])
            self.columns = self.input_columns[:2]  # Assume the first two columns are of interest
        elif self.input_columns is None:
            if self.columns is None:
                self.columns = list(self.df.columns[:2])  # Assume the first two columns are of interest
        else:
            raise ValueError("input_columns must be a list, dict, or None")

    def load_progress(self):
        pickle_files = [f for f in os.listdir(self.pickle_dir) if f.endswith('.pkl')]
        if pickle_files:
            latest_pickle = max(pickle_files, key=lambda x: os.path.getmtime(os.path.join(self.pickle_dir, x)))
            with open(os.path.join(self.pickle_dir, latest_pickle), 'rb') as f:
                self.id_mapping = pickle.load(f)
            self.logger.info(f"Loaded progress from {latest_pickle}")
        else:
            self.logger.info("No existing progress found. Starting from scratch.")

    def clear_all_progress(self):
        for file in os.listdir(self.pickle_dir):
            if file.endswith('.pkl'):
                os.remove(os.path.join(self.pickle_dir, file))
        self.logger.info("Cleared all progress files.")

    @staticmethod
    def get_ids(args):
        batch, source_type, dest_type, max_retries = args
        results = {}
        for id_ in batch:
            for attempt in range(max_retries):
                try:
                    if source_type == 'Gene_Name':
                        request = IdMappingClient.submit(source=source_type, dest=dest_type, ids={id_}, taxon_id='9606')
                    else:
                        request = IdMappingClient.submit(source=source_type, dest=dest_type, ids={id_})
                    # Add an exponential backoff with jitter to avoid thundering herd problem
                    time.sleep(min(1 ** attempt + random.uniform(0, 1), 10))
                    result = list(request.each_result())
                    results[id_] = [m['to'] for m in result] if result else []
                    break
                except Exception as e:
                    if attempt == max_retries - 1:
                        logging.error(f"Error processing {id_} after {max_retries} attempts: {str(e)}")
                        results[id_] = []
        return results

    def identify_untranslated_ids(self):
        untranslated_ids = set()
        source_column = f"{self.columns[0]}_{self.dest_type}"
        target_column = f"{self.columns[1]}_{self.dest_type}"

        # Check untranslated IDs in the source column
        untranslated_source = self.df[self.df[self.columns[0]] == self.df[source_column]][self.columns[0]].unique()
        untranslated_ids.update(untranslated_source)

        # Check untranslated IDs in the target column
        untranslated_target = self.df[self.df[self.columns[1]] == self.df[target_column]][self.columns[1]].unique()
        untranslated_ids.update(untranslated_target)

        self.logger.info(f"Found {len(untranslated_ids)} untranslated IDs to retry")

        return untranslated_ids

    def translate(self):
        self.load_data()
        self.load_progress()

        # calculate the ids that have not been translated yet
        remaining_ids = self.unique_ids - set(self.id_mapping.keys())
        id_batches = [list(remaining_ids)[i:i + self.batch_size] for i in range(0, len(remaining_ids), self.batch_size)]

        with Pool(self.processes) as pool:
            progress_bar = tqdm(total=len(id_batches), desc="Processing batches")
            for i, batch_results in enumerate(pool.imap_unordered(self.get_ids,
                                                                  [(batch, self.source_type, self.dest_type,
                                                                    self.max_retries)
                                                                   for batch in id_batches])):
                self.id_mapping.update(batch_results)
                progress_bar.update(1)
                self.logger.info(f"Processed batch {i + 1}/{len(id_batches)}")

                if (i + 1) % 2 == 0 or i == len(id_batches) - 1:
                    self.save_progress(f'checkpoint_batch_{i + 1}.pkl')

            progress_bar.close()

        self.apply_translation()

    def save_progress(self, filename: str):
        with open(os.path.join(self.pickle_dir, filename), 'wb') as f:
            pickle.dump(self.id_mapping, f)
        self.logger.info(f"Progress saved to {filename}")

    def apply_translation(self):
        self.logger.info("Applying translation to dataframe...")

        if self.multiple_mapping_strategy == 'expand':
            expanded_rows = []
            for _, row in self.df.iterrows():
                source_mappings = self.id_mapping.get(row[self.columns[0]])
                target_mappings = self.id_mapping.get(row[self.columns[1]])

                # If either mapping is None, skip this row
                if source_mappings is None or target_mappings is None:
                    continue

                # If mappings are empty lists, use the original identifiers
                source_mappings = source_mappings if source_mappings else [row[self.columns[0]]]
                target_mappings = target_mappings if target_mappings else [row[self.columns[1]]]

                for source, target in itertools.product(source_mappings, target_mappings):
                    new_row = row.copy()
                    new_row[f"{self.columns[0]}_{self.dest_type}"] = source
                    new_row[f"{self.columns[1]}_{self.dest_type}"] = target
                    expanded_rows.append(new_row)

            if expanded_rows:
                self.df = pd.DataFrame(expanded_rows)
            else:
                logging.warning("No valid translations found. The DataFrame is empty.")
        else:
            for column in self.columns:
                new_column = f"{column}_{self.dest_type}"
                self.df[new_column] = self.df[column].map(lambda x: self.id_mapping.get(x, [x]))
                self.df[new_column] = self.df[new_column].apply(lambda x: ';'.join(x) if isinstance(x, list) else x)

        # Save the translated dataframe
        self._save_dataframe(self.output_file)

    def run(self):
        self.logger.info(f"Starting ID translation process from {self.source_type} to {self.dest_type}")
        start_time = time.time()
        self.translate()

        untranslated_ids = self.identify_untranslated_ids()
        if untranslated_ids:
            self.logger.info(f"Retrying translation for {len(untranslated_ids)} untranslated IDs")
            self.translate_untranslated(untranslated_ids)

        end_time = time.time()
        self.logger.info(f"ID translation process completed in {end_time - start_time:.2f} seconds")
        self.analyze_results()

    def translate_untranslated(self, untranslated_ids):
        id_batches = [list(untranslated_ids)[i:i + self.batch_size] for i in
                      range(0, len(untranslated_ids), self.batch_size)]

        with Pool(self.processes) as pool:
            progress_bar = tqdm(total=len(id_batches), desc="Retrying untranslated batches")
            for i, batch_results in enumerate(pool.imap_unordered(self.get_ids,
                                                                  [(batch, self.source_type, self.dest_type,
                                                                    self.max_retries)
                                                                   for batch in id_batches])):
                self.id_mapping.update(batch_results)
                progress_bar.update(1)
                self.logger.info(f"Retried batch {i + 1}/{len(id_batches)}")

                if (i + 1) % 2 == 0 or i == len(id_batches) - 1:
                    self.save_progress(f'retry_checkpoint_batch_{i + 1}.pkl')

            progress_bar.close()

        self.apply_translation()

    def analyze_results(self):
        original_count = len(self.df)
        # Determine the translated count by checking if translated columns differ from original columns
        translated_columns = [f"{col}_{self.dest_type}" for col in self.columns]
        translated_count = self.df.apply(
            lambda row: all(
                row[orig_col] != row[trans_col] for orig_col, trans_col in zip(self.columns, translated_columns)),
            axis=1
        ).sum()
        expansion_factor = translated_count / original_count if original_count > 0 else 0

        self.logger.info(f"Original entry count: {original_count}")
        self.logger.info(f"Translated entry count: {translated_count}")
        self.logger.info(f"Expansion factor: {expansion_factor:.2f}")

        success_rate = (translated_count / original_count) * 100 if original_count > 0 else 0
        self.logger.info(f"Translation success rate: {success_rate:.2f}%")

    def remove_untranslated_entries(self, output_file: str = None):
        if self.df is None:
            raise ValueError("No data loaded. Please run the translation process first.")

        original_count = len(self.df)

        # Identify columns that contain translated IDs
        translated_columns = [f"{col}_{self.dest_type}" for col in self.columns]

        # Remove rows where any of the translated columns is identical to the original columns
        self.df = self.df[
            self.df.apply(
                lambda row: any(
                    row[orig_col] != row[trans_col] for orig_col, trans_col in zip(self.columns, translated_columns)),
                axis=1
            )
        ]

        removed_count = original_count - len(self.df)

        if output_file is None:
            file_name, file_extension = os.path.splitext(self.output_file)
            output_file = f"{file_name}_cleaned{file_extension}"

        # Save the cleaned dataframe
        self._save_dataframe(output_file)

        self.logger.info(f"Removed {removed_count} untranslated entries.")
        self.logger.info(f"Cleaned database saved to {output_file}")
        self.logger.info(f"Original entry count: {original_count}")
        self.logger.info(f"Cleaned entry count: {len(self.df)}")

        # Update the success rate
        self.analyze_results()

    def load_translated_dataframe(self, file_path: str):
        """
        Loads a translated DataFrame from a file.

        Args:
            file_path (str): The path to the file containing the translated DataFrame.

        Returns:
            pd.DataFrame: The loaded DataFrame.
        """
        file_extension = os.path.splitext(file_path)[1].lower()
        if file_extension == '.csv':
            self.df = pd.read_csv(file_path)
        elif file_extension in ['.xls', '.xlsx']:
            self.df = pd.read_excel(file_path)
        elif file_extension == '.tsv':
            self.df = pd.read_csv(file_path, sep='\t')
        else:
            raise ValueError(f"Unsupported file format: {file_extension}")
        self.logger.info(f"Loaded translated DataFrame from {file_path}")

    def translate_single_identifier(self, identifier: str, source_type: str, dest_type: str, taxon_id: int = None, sleep_time: int = 3,
                                    replace_in_db: bool = False, file_path: str = None):
        """
        Translates a single identifier and updates the DataFrame if the translation is successful.

        Args:
            identifier (str): The identifier to translate.
            source_type (str): The source type of the identifier.
            dest_type (str): The destination type for the translation.
            sleep_time (int): The time to wait (in seconds) for the request to complete.
            replace_in_db (bool): Flag to decide if the translation should replace the original in the database.
            file_path (str): Optional path to a file containing a pre-translated DataFrame to update.

        Returns:
            List[str]: The translated identifiers.
        """
        if file_path:
            self.load_translated_dataframe(file_path)

        translated_ids = []
        for attempt in range(self.max_retries):
            try:
                # Submit the request to the IdMappingClient
                if source_type == 'Gene_Name' and taxon_id:
                    request = IdMappingClient.submit(source=source_type, dest=dest_type, ids={identifier},
                                                     taxon_id=taxon_id)
                else:
                    request = IdMappingClient.submit(source=source_type, dest=dest_type, ids={identifier})

                # Wait for the request to complete
                time.sleep(sleep_time)

                # Process the result
                result = list(request.each_result())
                translated_ids = [m['to'] for m in result] if result else []
                break
            except Exception as e:
                if attempt == self.max_retries - 1:
                    self.logger.error(f"Error processing {identifier} after {self.max_retries} attempts: {str(e)}")

        # Automatically select the first translation
        if translated_ids:
            translated_ids = [translated_ids[0]]

        # Print the translated entry
        print(f"Original Identifier: {identifier}")
        print(f"Translated Identifier(s): {translated_ids if translated_ids else 'No translation found'}")

        # Optionally replace the identifier in the DataFrame
        if replace_in_db and translated_ids:
            source_column = f"{self.columns[0]}_{dest_type}"
            target_column = f"{self.columns[1]}_{dest_type}"

            for column, translated_column in zip(self.columns, [source_column, target_column]):
                self.df[translated_column] = self.df.apply(
                    lambda row: ';'.join(translated_ids) if row[column] == identifier else row[translated_column],
                    axis=1
                )

        return translated_ids

    def _save_dataframe(self, output_file):
        if self.df.empty:
            logging.warning("The DataFrame is empty. No file will be saved.")
            return

        file_extension = os.path.splitext(output_file)[1].lower()
        if file_extension == '.csv':
            self.df.to_csv(output_file, index=False)
        elif file_extension in ['.xls', '.xlsx']:
            self.df.to_excel(output_file, index=False)
        elif file_extension == '.tsv':
            self.df.to_csv(output_file, sep='\t', index=False)
        else:
            raise ValueError(f"Unsupported output file format: {file_extension}")

        self.logger.info(f"Results saved to {output_file}")

    def save_translated_dataframe(self, file_path: str):
        """
        Saves the translated DataFrame to a file.

        Args:
            file_path (str): The path where the translated DataFrame should be saved.
        """
        file_extension = os.path.splitext(file_path)[1].lower()
        if file_extension == '.csv':
            self.df.to_csv(file_path, index=False)
        elif file_extension in ['.xls', '.xlsx']:
            self.df.to_excel(file_path, index=False)
        elif file_extension == '.tsv':
            self.df.to_csv(file_path, sep='\t', index=False)
        else:
            raise ValueError(f"Unsupported file format: {file_extension}")
        self.logger.info(f"Translated DataFrame saved to {file_path}")




from __future__ import annotations

import re
import typing
from typing import Callable
import functools

import pypath_common._misc as _common

"""
Define, access and process fields from network inputs.
"""


def _to_fields(loc: dict) -> dict[str, Field]:

    return {k: Field(v) for k, v in loc.items() if k != 'cls'}


class NodeDef(typing.NamedTuple):
    id: Field
    id_type: Field
    entity_type: Field
    organism: Field

    def __new__(
            cls,
            id: Field,
            id_type: Field,
            entity_type: Field,
            organism: Field,
        ):

        args = _to_fields(locals())

        return super().__new__(cls, **args)


class EdgeDef(typing.NamedTuple):
    type: Field
    directed: Field
    effect: Field

    def __new__(cls, type, directed, effect):

        args = _to_fields(locals())

        return super().__new__(cls, **args)


class Field:

    _SEP0 = '?'
    _SEP1 = '!'
    _DEFVALS = (True, False, None)
    _SPECVAL = {'True': True, 'False': False, 'None': None}
    _RECOLPFX = re.compile(r'^__')

    def __init__(
            self,
            definition: str | Callable | tuple | dict,
        ):
        """
        Definition of a field in a data frame.

        Args:
            definition:
                Defines how to extract and process the value of the field from
                a data frame row (record). Most often it refers to a column,
                and defines additional processing steps. The following options
                are supported:
                - A string starting with "__": refers to a column, and it
                  can include further processing by using a simple
                  notation, see more at the ``short`` method of this class.
                - A string: a constant value (the same value will be
                  returned for all any record).
                - A string refering a method in a module, e.g.
                  "json.loads": this method will be used to process the
                  field. By default the function is called with complete
                  records, to operate on a single field instead, add the
                  column name in front, e.g. "__extra_attrs?json.loads".
                - Custom Python code: the code should evaluate into a
                  callable which then will be used to process the field.
                - More or less identical to the above, the definition can
                  be provided as a tuple.
                - A dict representation is also accepted, the dict should
                  contain the column name, the within field separator and
                  the mappings under the "col", "sep" and "maps" keys. If
                  "maps" is callable, it will be called on the field value.
        """

        self._setup(definition)


    def _setup(self, definition: str | Callable | tuple | dict):

        if isinstance(definition, str):

            if definition.startswith('__'):

                # special notation __ marks the string as column definition
                definition = (definition[2:],)

            elif re.match(r'^\w+$', definition):

                # directly provide a constant value
                self._access = lambda record: definition

            elif re.match(r'^[\.\w]+$', definition):

                # refer to any custom function from any module
                self._access = self._func_from_module(definition)

            else:

                # use custom python code
                definition = _common.code_to_func(definition)

        elif isinstance(definition, tuple):

            definition = self.short(self._SEP0.join(definition))

        if isinstance(definition, Callable):

            self._access = definition

        elif isinstance(definition, dict):

            self._access = functools.partial(
                self._access_by_param,
                param = definition,
            )


    @classmethod
    def short(cls, definition: str) -> dict:
        """
        Parse field processing parameters from a short string notation.

        The simple notation is specific to this module and makes it possible to
        easily define the most common, basic ways of processing data from
        columns. For the simplest cases not even this notation is required, for
        more complex cases you can always inject your custome Python code.

        The notation uses `?` and `!` as higher and a lower level separators,
        hence these characters can not be part of column names or separators
        used in the processed data or in the definition. The first part of the
        notation describes the column name, the separators used to split the
        column values and to split the values in the definition. The rest of
        the definition consists of potential mappings. The exact processing
        behaviour depends on the number and contents of the mappings:

        __column = Simply take the value of "column".
        __column!; = Split the value of "column" by ";".
        __column!;?Activation = Split the value of "column" and if the word
            "Activation" is among the values return True, otherwise False.
        __column!!|?Activation?Inhibition?Undefined|Unknown = Do not split the
            value of column, but split the values in this definition by "|".
            Then match the column value against the ones listed here, i.e.
            Activation = True, Inhibition = False, Undefined or Unknown = None.
            These (True, False, None) in this order are values of the default
            mapping, different values have to be specified explicitely.
        __column!!,?Ser!Serine,S,Ser?Tyr!Tyrosine,Tyr,Y?Thr,Threonine,Thr,T =
            Map single and three letter notation and full names of these amino
            acides to their three letter notation.

        In this function the "__" prefix is optional, as here we can be sure we
        refer to columns.
        In rare cases when the data is incompatible with `?` and `!` being used
        as separators, change the ``_SEP0`` and the ``_SEP1`` attributes of
        this class.
        """

        definition = _RECOLPFX.sub('', definition)
        col, *maps = definition.split(cls._SEP0)
        col, *seps = col.split(cls._SEP1)
        datasep, defsep = (s or None for s in (seps + [None] * 2)[:2])

        maps =
            self._func_from_module(maps)
                if maps and re.match(r'^[\w\.]+$', maps[0])
            [self._process_map(i, m, defsep) for i, m in enumerate(maps)]
        )

        return {
            'col': col,
            'sep': datasep,
            'maps': maps,
        }


    @classmethod
    def _process_map(
            cls,
            i: int,
            m: str,
            defsep: str | None = None,
        ) -> tuple[str | set[str], Any]:

        values, *map_to = m.split(cls._SEP1)[::-1] + [None]
        map_to = map_to[0] or cls._DEFVALS[i]
        map_to = cls._SPECVAL.get(map_to, map_to)
        values = set(values.split(defsep)) if defsep else values

        return values, map_to

    @classmethod
    def _access_by_param(
            cls,
            record: pd.DataFrame | dict,
            param: str | dict,
        ) -> Any:

        param = cls.short(param) if isinstance(param, str) else param
        return cls.access_by_param(record, param)


    @staticmethod
    def access_by_param(record: pd.DataFrame | dict, param: dict) -> Any:
        """
        Access the field value using processing parameters.

        Args:
            record:
                One row from a data frame or a dict.
            param:
                Processing parameters: a dict with keys "col", "sep" and
                "maps".

        Returns:
            The processed value of the field.
        """

        value = record[param['col']]
        op_one, op_many = value.__eq__, lambda x: x.__contains__(value)

        if param['sep']:

            value = set(value.split(param['sep']))
            op_one, op_many = value.__contains__, value.__and__

        if callable(param['maps']):

            return param['maps'](value)

        for v, map_to in param['maps']:

            if match := (op_many(v) if isinstance(v, set) else op_one(v)):

                return map_to

        return value


    @staticmethod
    def _access_column(col: str, record: pd.DataFrame | dict) -> Any:

        return record[col]


    @staticmethod
    def _func_from_module(func: str) -> Callable:

        # refer to any custom function from any module
        mod, *func = definition.rsplit('.', maxsplit = 1)
        mod = __import__(mod)
        return getattr(mod, func)


    def __call__(self, record: pd.DataFrame | dict) -> Any:

        return self._access(record)



#!/usr/bin/env python

#
# This file is part of the `neko` Python module
#
# Copyright 2023
# Insitut Curie
#
# File author(s): Marco Ruscone (marco.ruscone@curie.fr)
#
# Distributed under the GPLv3 license
# See the file `LICENSE` or read a copy at
# https://www.gnu.org/licenses/gpl-3.0.txt
#

"""
Package metadata (version, authors, etc).
"""

__all__ = ['get_metadata']

import os
import pathlib
import importlib.metadata

import toml

_VERSION = '0.0.1'


def get_metadata():
    """
    Basic package metadata.

    Retrieves package metadata from the current project directory or from
    the installed package.
    """

    here = pathlib.Path(__file__).parent
    pyproj_toml = 'pyproject.toml'
    meta = {}

    for project_dir in (here, here.parent):

        toml_path = str(project_dir.joinpath(pyproj_toml).absolute())

        if os.path.exists(toml_path):

            pyproject = toml.load(toml_path)

            meta = {
                'name': pyproject['tool']['poetry']['name'],
                'version': pyproject['tool']['poetry']['version'],
                'author': pyproject['tool']['poetry']['authors'],
                'license': pyproject['tool']['poetry']['license'],
                'full_metadata': pyproject,
            }

            break

    if not meta:

        try:

            meta = {
                k.lower(): v for k, v in
                importlib.metadata.metadata(here.name).items()
            }

        except importlib.metadata.PackageNotFoundError:

            pass

    meta['version'] = meta.get('version', None) or _VERSION

    return meta


metadata = get_metadata()
__version__ = metadata.get('version', None)
__author__ = metadata.get('author', None)
__license__ = metadata.get('license', None)



from typing import TYPE_CHECKING
import pandas as pd


def bool_col(df: pd.DataFrame, col: str, mappings: dict = None) -> pd.DataFrame:

    mappings = mappings or {
        "True": True,
        "False": False,
        "true": True,
        "false": False,
        1: True,
        -1: False,
        0: False,
    }

    if col not in df.columns:

        df[col] = False

    df[col] = df[col].replace(mappings).astype(bool)

    return df


def split_effect(
        df: pd.DataFrame,
        col: str = 'effect',
        mappings: dict = None,
        inhibition: bool = False,
    ) -> pd.DataFrame:
    """
    Split effects into two column notation (is_stimulation and is_inhibition).

    Parameters:
        col:
            Name of the column with the effect information.
        mapping:
            Mapping for the values to boolean.
        inhibition:
            Negate the effect: do it where True means inhibition.

    Returns:
        The data frame with col removed and is_stimulation and is_inhibition
        columns added.
    """

    df = bool_col(df, col, mappings)

    if inhibition:
        df[col] = ~df[col]

    df['is_inhibition'] = ~df[col]
    df.rename({col: 'is_stimulation'}, inplace = True, axis = 1)

    return df


def undirected_to_mutual(
        df: pd.DataFrame,
        source_col: str = 'source',
        target_col: str = 'target',
    ):

    cols = {source_col: target_col, target_col: source_col}

    df = (
        pd.concat([
            df,
            df.copy().rename(columns = cols),
        ]).
        drop_duplicates(
            subset = [source_col, target_col],
            ignore_index = True,
        )
    )

    return df



from graphviz import Digraph
from IPython.display import display
from yfiles_jupyter_graphs import GraphWidget
from typing import Dict


def wrap_node_name(node_name):
    if ":" in node_name:
        node_name = node_name.replace(":", "_")
    if node_name.startswith("COMPLEX"):
        return node_name[8:]
    else:
        return node_name


class NetworkVisualizer:
    """
    Visualize a network using Graphviz and yfiles.

    Args:
        - network (Network): A Network object.
        - predefined_node (str): The name of a node to display along with its connections.
        - color_by (str): The attribute to use for coloring nodes.
        - noi (bool): If True, only display the nodes of interest.

    Attributes:
        - dataframe_edges (DataFrame): DataFrame containing the edges of the network.
        - dataframe_nodes (DataFrame): DataFrame containing the nodes of the network.
        - initial_nodes (list): List of initial nodes.
        - color_by (str): The attribute to use for coloring nodes.
        - noi (bool): If True, only display the nodes of interest.
        - graph (Digraph): A Digraph object from the Graphviz library.
        - edge_colors (dict): Dictionary mapping edge effects to colors.
        - node_colors (dict): Dictionary mapping node names to colors.
        - predefined_node (str): The name of a node to display along with its connections.

    Methods:
        - set_custom_edge_colors(custom_edge_colors): Set custom edge colors.
        - set_node_colors(node_colors): Set custom node colors.
        - add_edges_to_graph(): Add edges to the graph.
        - add_nodes_to_graph(): Add nodes to the graph.
        - tissue_mapping(tissue_df): Color the nodes based on their expression in the tissue of interest.
        - render(output_file, view, highlight_nodes, highlight_color): Render the graph.
        - yfiles_visual(graph_layout, directed): Visualize the graph using yFiles.
        - vis_comparison(int_comparison, node_comparison, graph_layout, directed): Visualize the comparison of two networks.
    """
    def __init__(self, network, predefined_node=None, color_by="Effect", noi=False):
        net = network.copy()
        self.__dataframe_edges = net.convert_edgelist_into_genesymbol().copy()
        self.__dataframe_nodes = net.nodes
        self.initial_nodes = net.initial_nodes
        self.__color_by = color_by
        self.__noi = noi  # nodes of interest
        self.graph = Digraph(format='pdf')
        self.__edge_colors = {
            'stimulation': 'green',
            'inhibition': 'red',
            'form complex': 'blue',
            'bimodal': 'purple',  # Adding the mapping for "bimodal" effect
            'undefined': 'gray',  # Adding the mapping for "undefined" effect
            # Add more custom mappings if needed
        }
        self.__node_colors = {}  # Dictionary to store custom node colors
        self.predefined_node = wrap_node_name(predefined_node) if predefined_node else None
        # Apply wrap_node_name function to node names in dataframe_nodes
        self.__dataframe_nodes['Genesymbol'] = self.__dataframe_nodes['Genesymbol'].apply(wrap_node_name)
        self.__dataframe_nodes['Uniprot'] = self.__dataframe_nodes['Uniprot'].apply(wrap_node_name)

        # Apply wrap_node_name function to node names in "source" and "target" columns of dataframe_edges
        self.__dataframe_edges['source'] = self.__dataframe_edges['source'].apply(wrap_node_name)
        self.__dataframe_edges['target'] = self.__dataframe_edges['target'].apply(wrap_node_name)

        self.__dataframe_edges = self.__dataframe_edges.drop_duplicates(subset=['source', 'target', 'Type', 'Effect'])

        self.__add_edges_to_graph()
        self.__add_nodes_to_graph()

    def __set_custom_edge_colors(self, custom_edge_colors):
        # Update the edge_colors dictionary with custom mappings
        self.__edge_colors.update(custom_edge_colors)

    def __set_node_colors(self, node_colors):
        # Update the node_colors dictionary with custom node colorsdataframe_nodes
        self.__node_colors.update(node_colors)

    def __add_edges_to_graph(self):
        for _, row in self.__dataframe_edges.iterrows():
            effect = row['Effect']
            source = wrap_node_name(row['source'])
            target = wrap_node_name(row['target'])

            # Display only edges connected to the predefined node
            if self.predefined_node and (source != self.predefined_node and target != self.predefined_node):
                continue

            # Determine edge attributes based on effect
            if effect == 'stimulation':
                arrowhead = 'normal'
                color = 'green'
                dir = 'forward'
            elif effect == 'inhibition':
                arrowhead = 'tee'
                color = 'red'
                dir = 'forward'
            elif effect == 'form complex':
                arrowhead = 'dot'
                color = 'blue'
                dir = 'forward'
            elif effect == 'bimodal':
                arrowhead = 'diamond'
                color = 'purple'
                dir = 'forward'
            elif effect == 'undefined':
                arrowhead = 'normal'  # Default arrow
                color = 'black'
                dir = 'forward'
            else:
                arrowhead = 'normal'  # Default arrow
                color = 'black'
                dir = 'none'

            # Add the edge to the graph with specified attributes
            self.graph.edge(source, target, color=color, arrowhead=arrowhead, dir=dir)

    def __add_nodes_to_graph(self):
        for _, row in self.__dataframe_nodes.iterrows():
            node = row['Genesymbol']
            # add function to set color
            node_color = 'lightgray'
            if node in self.initial_nodes and self.__noi:
                node_color = 'lightyellow'
            node_color = self.__node_colors.get(node, node_color)

            # Display only the predefined node and its connections
            if self.predefined_node and (node != self.predefined_node):
                continue

            wrapped_node = wrap_node_name(node)
            self.graph.node(wrapped_node, style='filled', fillcolor=node_color)

    def tissue_mapping(self, tissue_df):
        """
        Color the nodes based on their expression in the tissue of interest (based on data from The Human Protein Atlas).

        Args:
            tissue_df (DataFrame): DataFrame containing results indicating whether each gene symbol has tissue annotations containing the selected tissue.
        """
        for _, row in tissue_df.iterrows():
            gene_symbol = row['Genesymbol']
            in_tissue = row['in_tissue']
            node_color = 'lightblue' if in_tissue else 'lightgray'
            self.__node_colors[gene_symbol] = node_color

    def render(self, output_file='network', view=False, highlight_nodes=None, highlight_color='lightyellow'):
        """
        Render the graph.

        Args:
            output_file (str): The name of the output file.
            view (bool): If True, display the graph.
            highlight_nodes (list): List of nodes to highlight.
            highlight_color (str): Color to use for highlighting nodes.
        """
        # If highlight_nodes is provided, set the color for each node in the list
        if highlight_nodes is not None:
            for node in highlight_nodes:
                # first check that the node is in the node dataframe
                if wrap_node_name(node) in self.__dataframe_nodes['Genesymbol'].values:
                    # then change the color only of the node in the graph
                    self.graph.node(wrap_node_name(node), style='filled', fillcolor=highlight_color)
        if view:
            self.graph.view(filename=output_file)
        else:
            self.graph.render(filename=output_file)
            display(self.graph)  # Display the graph directly in the Jupyter Notebook

    def yfiles_visual(
        self,
        graph_layout,
        directed,
    ):
        # creating empty object for visualization
        w = GraphWidget()

        # filling w with nodes
        objects = []
        for idx, item in self.__dataframe_nodes.iterrows():
            obj = {
                "id": self.__dataframe_nodes["Uniprot"].loc[idx],
                "properties": {"label": self.__dataframe_nodes["Genesymbol"].loc[idx]},
                "color": "#ffffff",
                "styles": {"backgroundColor": "#ffffff"}
            }
            objects.append(obj)
        w.nodes = objects

        # filling w with edges
        objects = []
        for index, row in self.__dataframe_edges.iterrows():
            obj = {
                "id": self.__dataframe_edges["Effect"].loc[index],
                "start": self.__dataframe_edges["source"].loc[index],
                "end": self.__dataframe_edges["target"].loc[index],
                "properties": {"references": self.__dataframe_edges["References"].loc[index]}}
            objects.append(obj)
        w.edges = objects

        def custom_edge_color_mapping(edge: Dict):
            """let the edge be red if the interaction is an inhibition, else green"""
            return "#fa1505" if edge['id'] == "inhibition" else "#05e60c"

        w.set_edge_color_mapping(custom_edge_color_mapping)

        def custom_node_color_mapping(node: Dict):
            return {"color": "#ffffff"}

        w.set_node_styles_mapping(custom_node_color_mapping)

        def custom_factor_mapping(node: Dict):
            """choose random factor"""
            return 5

        w.set_node_scale_factor_mapping(custom_factor_mapping)

        def custom_label_styles_mapping(node: Dict):
            """let the label be the negated purple big index"""
            return {
                'text': node["properties"]["label"],
                'backgroundColor': None,
                'fontSize': 40,
                'color': '#030200',
                'shape': 'round-rectangle',
                'textAlignment': 'center'
            }

        w.set_node_label_mapping(custom_label_styles_mapping)

        w.directed = directed
        w.graph_layout = graph_layout

        display(w)

    def vis_comparison(
        self,
        int_comparison,
        node_comparison,
        graph_layout,
        directed,
    ):
        # creating empty object for visualization
        w = GraphWidget()

        objects = []
        for idx, item in node_comparison.iterrows():
            obj = {
                "id": node_comparison["node"].loc[idx],
                "properties": {"label": node_comparison["node"].loc[idx],
                               "comparison": node_comparison["comparison"].loc[idx], },
                "color": "#ffffff",
                #       "styles":{"backgroundColor":"#ffffff"}
            }
            objects.append(obj)
        w.nodes = objects

        # filling w with edges
        objects = []
        for index, row in int_comparison.iterrows():
            obj = {
                "id": int_comparison["comparison"].loc[index],
                "properties": {
                    "comparison": int_comparison["comparison"].loc[index]},
                "start": int_comparison["source"].loc[index],
                "end": int_comparison["target"].loc[index]
            }
            objects.append(obj)
        w.edges = objects

        def custom_node_color_mapping(node: Dict):
            if node['properties']['comparison'] == "Unique to Network 1":
                return {"color": "#f5f536"}
            elif node['properties']['comparison'] == "Unique to Network 2":
                return {"color": "#36f55f"}
            elif node['properties']['comparison'] == "Common":
                return {"color": "#3643f5"}

        w.set_node_styles_mapping(custom_node_color_mapping)

        def custom_factor_mapping(node: Dict):
            """choose random factor"""
            return 5

        w.set_node_scale_factor_mapping(custom_factor_mapping)

        def custom_label_styles_mapping(node: Dict):
            """let the label be the negated purple big index"""
            return {
                'text': node["id"],
                'backgroundColor': None,
                'fontSize': 20,
                'color': '#030200',
                'position': 'center',
                'maximumWidth': 130,
                'wrapping': 'word',
                'textAlignment': 'center'
            }

        w.set_node_label_mapping(custom_label_styles_mapping)

        def custom_edge_color_mapping(edge: Dict):
            if edge['id'] == "Unique to Network 1":
                return "#e3941e"
            elif edge['id'] == "Unique to Network 2":
                return "#36f55f"
            elif edge['id'] == "Common":
                return "#3643f5"
            elif edge['id'] == "Conflicting":
                return "#ffcc00"

        w.set_edge_color_mapping(custom_edge_color_mapping)

        w.directed = directed
        w.graph_layout = graph_layout

        display(w)
